# Copyright 2010-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from _emerge.AtomArg import AtomArg
from _emerge.Package import Package
from _emerge.PackageArg import PackageArg
from _emerge.UseFlagDisplay import pkg_use_display

from portage.dep import check_required_use
from portage.output import colorize
from portage._sets.base import InternalPackageSet
from portage.util import writemsg
from portage.versions import cpv_getversion, vercmp


class slot_conflict_handler:
    """This class keeps track of all slot conflicts and provides
    an interface to get possible solutions.

    How it works:
    If two packages have been pulled into a slot, one needs to
    go away. This class focuses on cases where this can be achieved
    with a change in USE settings.

    1) Find out what causes a given slot conflict. There are
    three possibilities:

            a) One parent needs foo-1:0 and another one needs foo-2:0,
            nothing we can do about this. This is called a 'version
            based conflict'.

            b) All parents of one of the conflict packages could use
            another conflict package. This is called an 'unspecific
            conflict'. This should be caught by the backtracking logic.
            Ask the user to enable -uN (if not already enabled). If -uN is
            enabled, this case is treated in the same way as c).

            c) Neither a 'version based conflict' nor an 'unspecific
            conflict'. Ignoring use deps would result in an
            'unspecific conflict'. This is called a 'specific conflict'.
            This is the only conflict we try to find suggestions for.

    2) Computing suggestions.

    Def.: "configuration": A list of packages, containing exactly one
                    package from each slot conflict.

    We try to find USE changes such that all parents of conflict packages
    can work with a package in the configuration we're looking at. This
    is done for all possible configurations, except if the 'all-ebuild'
    configuration has a suggestion. In this case we immediately abort the
    search.
    For the current configuration, all use flags that are part of violated
    use deps are computed. This is done for every slot conflict on its own.

    Def.: "solution (candidate)": An assignment of "enabled" / "disabled"
                    values for the use flags that are part of violated use deps.

    Now all involved use flags for the current configuration are known. For
    now they have an undetermined value. Fix their value in the
    following cases:
            * The use dep in the parent atom is unconditional.
            * The parent package is 'installed'.
            * The conflict package is 'installed'.

    USE of 'installed' packages can't be changed. This always requires a
    non-installed package.

    During this procedure, contradictions may occur. In this case the
    configuration has no solution.

    Now generate all possible solution candidates with fixed values. Check
    if they don't introduce new conflicts.

    We have found a valid assignment for all involved use flags. Compute
    the needed USE changes and prepare the message for the user.
    """

    _check_configuration_max = 1024

    def __init__(self, depgraph):
        self.depgraph = depgraph
        self.myopts = depgraph._frozen_config.myopts
        self.debug = "--debug" in self.myopts
        if self.debug:
            writemsg("Starting slot conflict handler\n", noiselevel=-1)

        # List of tuples, where each tuple represents a slot conflict.
        self.all_conflicts = []
        for conflict in depgraph._dynamic_config._package_tracker.slot_conflicts():
            self.all_conflicts.append((conflict.root, conflict.atom, conflict.pkgs))

        # A dict mapping packages to pairs of parent package
        # and parent atom
        self.all_parents = depgraph._dynamic_config._parent_atoms

        # set containing all nodes that are part of a slot conflict
        conflict_nodes = set()

        # a list containing list of packages that form a slot conflict
        conflict_pkgs = []

        # a list containing sets of (parent, atom) pairs that have pulled packages
        # into the same slot
        all_conflict_atoms_by_slotatom = []

        # fill conflict_pkgs, all_conflict_atoms_by_slotatom
        for root, atom, pkgs in self.all_conflicts:
            conflict_pkgs.append(list(pkgs))
            all_conflict_atoms_by_slotatom.append(set())

            for pkg in pkgs:
                conflict_nodes.add(pkg)
                for ppkg, atom in self.all_parents.get(pkg):
                    all_conflict_atoms_by_slotatom[-1].add((ppkg, atom))

        # Variable that holds the non-explanation part of the message.
        self.conflict_msg = []
        # If any conflict package was pulled in only by unspecific atoms, then
        # the user forgot to enable --newuse and/or --update.
        self.conflict_is_unspecific = False

        # Indicate if the conflict is caused by incompatible version requirements
        # cat/pkg-2 pulled in, but a parent requires <cat/pkg-2
        self.is_a_version_conflict = False

        self._prepare_conflict_msg_and_check_for_specificity()

        # a list of dicts that holds the needed USE values to solve all conflicts
        self.solutions = []

        # a list of dicts that holds the needed USE changes to solve all conflicts
        self.changes = []

        # configuration = a list of packages with exactly one package from every
        # single slot conflict
        config_gen = _configuration_generator(conflict_pkgs)
        first_config = True

        # Go through all configurations and collect solutions
        while True:
            config = config_gen.get_configuration()
            if not config:
                break

            if self.debug:
                writemsg("\nNew configuration:\n", noiselevel=-1)
                for pkg in config:
                    writemsg("   {}\n".format(pkg), noiselevel=-1)
                writemsg("\n", noiselevel=-1)

            new_solutions = self._check_configuration(
                config, all_conflict_atoms_by_slotatom, conflict_nodes
            )

            if new_solutions:
                self.solutions.extend(new_solutions)

                if first_config:
                    # If the "all ebuild"-config gives a solution, use it.
                    # Otherwise enumerate all other solutions.
                    if self.debug:
                        writemsg(
                            "All-ebuild configuration has a solution. Aborting search.\n",
                            noiselevel=-1,
                        )
                    break
            first_config = False

            if len(conflict_pkgs) > 4:
                # The number of configurations to check grows exponentially in the number of conflict_pkgs.
                # To prevent excessive running times, only check the "all-ebuild" configuration,
                # if the number of conflict packages is too large.
                if self.debug:
                    writemsg(
                        "\nAborting search due to excessive number of configurations.\n",
                        noiselevel=-1,
                    )
                break

        for solution in self.solutions:
            self._add_change(self._get_change(solution))

    def get_conflict(self):
        return "".join(self.conflict_msg)

    def _is_subset(self, change1, change2):
        """
        Checks if a set of changes 'change1' is a subset of the changes 'change2'.
        """
        # All pkgs of change1 have to be in change2.
        # For every package in change1, the changes have to be a subset of
        # the corresponding changes in change2.
        for pkg in change1:
            if pkg not in change2:
                return False

            for flag in change1[pkg]:
                if flag not in change2[pkg]:
                    return False
                if change1[pkg][flag] != change2[pkg][flag]:
                    return False
        return True

    def _add_change(self, new_change):
        """
        Make sure to keep only minimal changes. If "+foo", does the job, discard "+foo -bar".
        """
        changes = self.changes
        # Make sure there is no other solution that is a subset of the new solution.
        ignore = False
        to_be_removed = []
        for change in changes:
            if self._is_subset(change, new_change):
                ignore = True
                break
            elif self._is_subset(new_change, change):
                to_be_removed.append(change)

        if not ignore:
            # Discard all existing changes that are a superset of the new change.
            for obsolete_change in to_be_removed:
                changes.remove(obsolete_change)
            changes.append(new_change)

    def _get_change(self, solution):
        _pkg_use_enabled = self.depgraph._pkg_use_enabled
        new_change = {}
        for pkg in solution:
            for flag, state in solution[pkg].items():
                flag = pkg.iuse.get_flag(flag)
                if flag is None:
                    # Triggered by use-dep defaults.
                    continue
                if state == "enabled" and flag not in _pkg_use_enabled(pkg):
                    new_change.setdefault(pkg, {})[flag] = True
                elif state == "disabled" and flag in _pkg_use_enabled(pkg):
                    new_change.setdefault(pkg, {})[flag] = False
        return new_change

    def _prepare_conflict_msg_and_check_for_specificity(self):
        """
        Print all slot conflicts in a human readable way.
        """
        _pkg_use_enabled = self.depgraph._pkg_use_enabled
        usepkgonly = "--usepkgonly" in self.myopts
        need_rebuild = {}
        verboseconflicts = "--verbose-conflicts" in self.myopts
        any_omitted_parents = False
        msg = self.conflict_msg
        indent = "  "
        msg.append(
            "\n!!! Multiple package instances within a single "
            + "package slot have been pulled\n"
        )
        msg.append(
            "!!! into the dependency graph, resulting" + " in a slot conflict:\n\n"
        )

        for root, slot_atom, pkgs in self.all_conflicts:
            msg.append("{}".format(slot_atom))
            if root != self.depgraph._frozen_config._running_root.root:
                msg.append(" for {}".format(root))
            msg.append("\n\n")

            for pkg in pkgs:
                msg.append(indent)
                msg.append(
                    "%s %s"
                    % (
                        pkg,
                        pkg_use_display(
                            pkg,
                            self.depgraph._frozen_config.myopts,
                            modified_use=self.depgraph._pkg_use_enabled(pkg),
                        ),
                    )
                )
                parent_atoms = self.all_parents.get(pkg)
                if parent_atoms:
                    # Create a list of collision reasons and map them to sets
                    # of atoms.
                    # Possible reasons:
                    # 	("version", "ge") for operator >=, >
                    # 	("version", "eq") for operator =, ~
                    # 	("version", "le") for operator <=, <
                    # 	("use", "<some use flag>") for unmet use conditionals
                    collision_reasons = {}
                    num_all_specific_atoms = 0

                    for ppkg, atom in parent_atoms:
                        if not atom.soname:
                            atom_set = InternalPackageSet(initial_atoms=(atom,))
                            atom_without_use_set = InternalPackageSet(
                                initial_atoms=(atom.without_use,)
                            )
                            atom_without_use_and_slot_set = InternalPackageSet(
                                initial_atoms=(atom.without_use.without_slot,)
                            )

                        for other_pkg in pkgs:
                            if other_pkg == pkg:
                                continue

                            if atom.soname:
                                # The soname does not match.
                                key = ("soname", atom)
                                atoms = collision_reasons.get(key, set())
                                atoms.add((ppkg, atom, other_pkg))
                                num_all_specific_atoms += 1
                                collision_reasons[key] = atoms
                            elif not atom_without_use_and_slot_set.findAtomForPackage(
                                other_pkg, modified_use=_pkg_use_enabled(other_pkg)
                            ):
                                if atom.operator is not None:
                                    # The version range does not match.
                                    sub_type = None
                                    if atom.operator in (">=", ">"):
                                        sub_type = "ge"
                                    elif atom.operator in ("=", "~"):
                                        sub_type = "eq"
                                    elif atom.operator in ("<=", "<"):
                                        sub_type = "le"

                                    key = ("version", sub_type)
                                    atoms = collision_reasons.get(key, set())
                                    atoms.add((ppkg, atom, other_pkg))
                                    num_all_specific_atoms += 1
                                    collision_reasons[key] = atoms

                            elif not atom_without_use_set.findAtomForPackage(
                                other_pkg, modified_use=_pkg_use_enabled(other_pkg)
                            ):
                                # The slot and/or sub_slot does not match.
                                key = (
                                    "slot",
                                    (atom.slot, atom.sub_slot, atom.slot_operator),
                                )
                                atoms = collision_reasons.get(key, set())
                                atoms.add((ppkg, atom, other_pkg))
                                num_all_specific_atoms += 1
                                collision_reasons[key] = atoms

                            elif not atom_set.findAtomForPackage(
                                other_pkg, modified_use=_pkg_use_enabled(other_pkg)
                            ):
                                missing_iuse = other_pkg.iuse.get_missing_iuse(
                                    atom.unevaluated_atom.use.required
                                )
                                if missing_iuse:
                                    for flag in missing_iuse:
                                        atoms = collision_reasons.get(
                                            ("use", flag), set()
                                        )
                                        atoms.add((ppkg, atom, other_pkg))
                                        collision_reasons[("use", flag)] = atoms
                                    num_all_specific_atoms += 1
                                else:
                                    # Use conditionals not met.
                                    violated_atom = atom.violated_conditionals(
                                        _pkg_use_enabled(other_pkg),
                                        other_pkg.iuse.is_valid_flag,
                                    )
                                    if violated_atom.use is None:
                                        # Something like bug #453400 caused the
                                        # above findAtomForPackage call to
                                        # return None unexpectedly.
                                        msg = (
                                            "\n\n!!! BUG: Detected "
                                            "USE dep match inconsistency:\n"
                                            "\tppkg: %s\n"
                                            "\tviolated_atom: %s\n"
                                            "\tatom: %s unevaluated: %s\n"
                                            "\tother_pkg: %s IUSE: %s USE: %s\n"
                                            % (
                                                ppkg,
                                                violated_atom,
                                                atom,
                                                atom.unevaluated_atom,
                                                other_pkg,
                                                sorted(other_pkg.iuse.all),
                                                sorted(_pkg_use_enabled(other_pkg)),
                                            )
                                        )
                                        writemsg(msg, noiselevel=-2)
                                        raise AssertionError(
                                            "BUG: USE dep match inconsistency"
                                        )
                                    for flag in violated_atom.use.enabled.union(
                                        violated_atom.use.disabled
                                    ):
                                        atoms = collision_reasons.get(
                                            ("use", flag), set()
                                        )
                                        atoms.add((ppkg, atom, other_pkg))
                                        collision_reasons[("use", flag)] = atoms
                                    num_all_specific_atoms += 1
                            elif isinstance(ppkg, AtomArg) and other_pkg.installed:
                                parent_atoms = collision_reasons.get(
                                    ("AtomArg", None), set()
                                )
                                parent_atoms.add((ppkg, atom))
                                collision_reasons[("AtomArg", None)] = parent_atoms
                                num_all_specific_atoms += 1

                    msg.append(" pulled in by\n")

                    selected_for_display = set()
                    unconditional_use_deps = set()

                    for (ctype, sub_type), parents in collision_reasons.items():
                        # From each (ctype, sub_type) pair select at least one atom.
                        # Try to select as few atoms as possible

                        if ctype == "version":
                            # Find the atom with version that is as far away as possible.
                            best_matches = {}
                            for ppkg, atom, other_pkg in parents:
                                if atom.cp in best_matches:
                                    cmp = vercmp(
                                        cpv_getversion(atom.cpv),
                                        cpv_getversion(best_matches[atom.cp][1].cpv),
                                    )

                                    if (
                                        (sub_type == "ge" and cmp > 0)
                                        or (sub_type == "le" and cmp < 0)
                                        or (sub_type == "eq" and cmp > 0)
                                    ):
                                        best_matches[atom.cp] = (ppkg, atom)
                                else:
                                    best_matches[atom.cp] = (ppkg, atom)
                                if verboseconflicts:
                                    selected_for_display.add((ppkg, atom))
                            if not verboseconflicts:
                                selected_for_display.update(best_matches.values())
                        elif ctype in ("soname", "slot"):
                            # Check for packages that might need to
                            # be rebuilt, but cannot be rebuilt for
                            # some reason.
                            for ppkg, atom, other_pkg in parents:
                                if not (isinstance(ppkg, Package) and ppkg.installed):
                                    continue
                                if not (atom.soname or atom.slot_operator_built):
                                    continue
                                if self.depgraph._frozen_config.excluded_pkgs.findAtomForPackage(
                                    ppkg,
                                    modified_use=self.depgraph._pkg_use_enabled(ppkg),
                                ):
                                    selected_for_display.add((ppkg, atom))
                                    need_rebuild[ppkg] = "matched by --exclude argument"
                                elif self.depgraph._frozen_config.useoldpkg_atoms.findAtomForPackage(
                                    ppkg,
                                    modified_use=self.depgraph._pkg_use_enabled(ppkg),
                                ):
                                    selected_for_display.add((ppkg, atom))
                                    need_rebuild[
                                        ppkg
                                    ] = "matched by --useoldpkg-atoms argument"
                                elif usepkgonly:
                                    # This case is tricky, so keep quiet in order to avoid false-positives.
                                    pass
                                elif not self.depgraph._equiv_ebuild_visible(ppkg):
                                    selected_for_display.add((ppkg, atom))
                                    need_rebuild[
                                        ppkg
                                    ] = "ebuild is masked or unavailable"

                            for ppkg, atom, other_pkg in parents:
                                selected_for_display.add((ppkg, atom))
                                if not verboseconflicts:
                                    break
                        elif ctype == "use":
                            # Prefer atoms with unconditional use deps over, because it's
                            # not possible to change them on the parent, which means there
                            # are fewer possible solutions.
                            use = sub_type
                            for ppkg, atom, other_pkg in parents:
                                missing_iuse = other_pkg.iuse.get_missing_iuse(
                                    atom.unevaluated_atom.use.required
                                )
                                if missing_iuse:
                                    unconditional_use_deps.add((ppkg, atom))
                                else:
                                    parent_use = None
                                    if isinstance(ppkg, Package):
                                        parent_use = _pkg_use_enabled(ppkg)
                                    violated_atom = (
                                        atom.unevaluated_atom.violated_conditionals(
                                            _pkg_use_enabled(other_pkg),
                                            other_pkg.iuse.is_valid_flag,
                                            parent_use=parent_use,
                                        )
                                    )
                                    # It's possible for autounmask to change
                                    # parent_use such that the unevaluated form
                                    # of the atom now matches, even though the
                                    # earlier evaluated form (from before
                                    # autounmask changed parent_use) does not.
                                    # In this case (see bug #374423), it's
                                    # expected that violated_atom.use is None.
                                    # Since the atom now matches, we don't want
                                    # to display it in the slot conflict
                                    # message, so we simply ignore it and rely
                                    # on the autounmask display to communicate
                                    # the necessary USE change to the user.
                                    if violated_atom.use is None:
                                        continue
                                    if (
                                        use in violated_atom.use.enabled
                                        or use in violated_atom.use.disabled
                                    ):
                                        unconditional_use_deps.add((ppkg, atom))
                                # When USE flags are removed, it can be
                                # essential to see all broken reverse
                                # dependencies here, so don't omit any.
                                # If the list is long, people can simply
                                # use a pager.
                                selected_for_display.add((ppkg, atom))
                        elif ctype == "AtomArg":
                            for ppkg, atom in parents:
                                selected_for_display.add((ppkg, atom))

                    def highlight_violations(atom, version, use, slot_violated):
                        """Colorize parts of an atom"""
                        atom_str = "{}".format(atom)
                        colored_idx = set()
                        if version:
                            op = atom.operator
                            ver = None
                            if atom.cp != atom.cpv:
                                ver = cpv_getversion(atom.cpv)
                            slot = atom.slot
                            sub_slot = atom.sub_slot
                            slot_operator = atom.slot_operator

                            if op == "=*":
                                op = "="
                                ver += "*"

                            slot_str = ""
                            if slot:
                                slot_str = ":" + slot
                            if sub_slot:
                                slot_str += "/" + sub_slot
                            if slot_operator:
                                slot_str += slot_operator

                            # Compute color_idx before adding the color codes
                            # as these change the indices of the letters.
                            if op is not None:
                                colored_idx.update(range(len(op)))

                            if ver is not None:
                                start = atom_str.rfind(ver)
                                end = start + len(ver)
                                colored_idx.update(range(start, end))

                            if slot_str:
                                ii = atom_str.find(slot_str)
                                colored_idx.update(range(ii, ii + len(slot_str)))

                            if op is not None:
                                atom_str = atom_str.replace(op, colorize("BAD", op), 1)

                            if ver is not None:
                                start = atom_str.rfind(ver)
                                end = start + len(ver)
                                atom_str = (
                                    atom_str[:start]
                                    + colorize("BAD", ver)
                                    + atom_str[end:]
                                )

                            if slot_str:
                                atom_str = atom_str.replace(
                                    slot_str, colorize("BAD", slot_str), 1
                                )

                        elif slot_violated:
                            slot = atom.slot
                            sub_slot = atom.sub_slot
                            slot_operator = atom.slot_operator

                            slot_str = ""
                            if slot:
                                slot_str = ":" + slot
                            if sub_slot:
                                slot_str += "/" + sub_slot
                            if slot_operator:
                                slot_str += slot_operator

                            if slot_str:
                                ii = atom_str.find(slot_str)
                                colored_idx.update(range(ii, ii + len(slot_str)))
                                atom_str = atom_str.replace(
                                    slot_str, colorize("BAD", slot_str), 1
                                )

                        if use and atom.use.tokens:
                            use_part_start = atom_str.find("[")
                            use_part_end = atom_str.find("]")

                            new_tokens = []
                            # Compute start index in non-colored atom.
                            ii = str(atom).find("[") + 1
                            for token in atom.use.tokens:
                                if token.lstrip("-!").rstrip("=?") in use:
                                    new_tokens.append(colorize("BAD", token))
                                    colored_idx.update(range(ii, ii + len(token)))
                                else:
                                    new_tokens.append(token)
                                ii += 1 + len(token)

                            atom_str = (
                                atom_str[:use_part_start]
                                + "[{}]".format(",".join(new_tokens))
                                + atom_str[use_part_end + 1 :]
                            )

                        return atom_str, colored_idx

                    # Show unconditional use deps first, since those
                    # are more problematic than the conditional kind.
                    ordered_list = list(unconditional_use_deps)
                    if len(selected_for_display) > len(unconditional_use_deps):
                        for parent_atom in selected_for_display:
                            if parent_atom not in unconditional_use_deps:
                                ordered_list.append(parent_atom)
                    for parent_atom in ordered_list:
                        parent, atom = parent_atom
                        if isinstance(parent, Package):
                            use_display = pkg_use_display(
                                parent,
                                self.depgraph._frozen_config.myopts,
                                modified_use=self.depgraph._pkg_use_enabled(parent),
                            )
                        else:
                            use_display = ""
                        if atom.soname:
                            msg.append(
                                "{} required by {} {}\n".format(
                                    atom, parent, use_display
                                )
                            )
                        elif isinstance(parent, PackageArg):
                            # For PackageArg it's
                            # redundant to display the atom attribute.
                            msg.append("{}\n".format(parent))
                        elif isinstance(parent, AtomArg):
                            msg.append(2 * indent)
                            msg.append("{} (Argument)\n".format(atom))
                        else:
                            # Display the specific atom from SetArg or
                            # Package types.
                            version_violated = False
                            slot_violated = False
                            use = []
                            for (ctype, sub_type), parents in collision_reasons.items():
                                for x in parents:
                                    if parent == x[0] and atom == x[1]:
                                        if ctype == "version":
                                            version_violated = True
                                        elif ctype == "slot":
                                            slot_violated = True
                                        elif ctype == "use":
                                            use.append(sub_type)
                                        break

                            atom_str, colored_idx = highlight_violations(
                                atom.unevaluated_atom,
                                version_violated,
                                use,
                                slot_violated,
                            )

                            if version_violated or slot_violated:
                                self.is_a_version_conflict = True

                            cur_line = "{} required by {} {}\n".format(
                                atom_str,
                                parent,
                                use_display,
                            )
                            marker_line = ""
                            for ii in range(len(cur_line)):
                                if ii in colored_idx:
                                    marker_line += "^"
                                else:
                                    marker_line += " "
                            marker_line += "\n"
                            msg.append(2 * indent)
                            msg.append(cur_line)
                            msg.append(2 * indent)
                            msg.append(marker_line)

                    if not selected_for_display:
                        msg.append(2 * indent)
                        msg.append(
                            "(no parents that aren't satisfied by other packages in this slot)\n"
                        )
                        self.conflict_is_unspecific = True

                    omitted_parents = num_all_specific_atoms - len(selected_for_display)
                    if omitted_parents:
                        any_omitted_parents = True
                        msg.append(2 * indent)
                        if len(selected_for_display) > 1:
                            msg.append(
                                "(and %d more with the same problems)\n"
                                % omitted_parents
                            )
                        else:
                            msg.append(
                                "(and %d more with the same problem)\n"
                                % omitted_parents
                            )
                else:
                    msg.append(" (no parents)\n")
                msg.append("\n")

        if any_omitted_parents:
            msg.append(
                colorize(
                    "INFORM",
                    "NOTE: Use the '--verbose-conflicts'"
                    " option to display parents omitted above",
                )
            )
            msg.append("\n")

        if need_rebuild:
            msg.append(
                "\n!!! The slot conflict(s) shown above involve package(s) which may need to\n"
            )
            msg.append(
                "!!! be rebuilt in order to solve the conflict(s). However, the following\n"
            )
            msg.append("!!! package(s) cannot be rebuilt for the reason(s) shown:\n\n")
            for ppkg, reason in need_rebuild.items():
                msg.append("{}{}: {}\n".format(indent, ppkg, reason))
            msg.append("\n")

        msg.append("\n")

    def get_explanation(self):
        msg = ""

        if self.is_a_version_conflict:
            return None

        if self.conflict_is_unspecific and not (
            "--newuse" in self.myopts and "--update" in self.myopts
        ):
            msg += "!!! Enabling --newuse and --update might solve this conflict.\n"
            msg += "!!! If not, it might help emerge to give a more specific suggestion.\n\n"
            return msg

        solutions = self.solutions
        if not solutions:
            return None

        if len(solutions) == 1:
            if len(self.all_conflicts) == 1:
                msg += "It might be possible to solve this slot collision\n"
            else:
                msg += "It might be possible to solve these slot collisions\n"
            msg += "by applying all of the following changes:\n"
        else:
            if len(self.all_conflicts) == 1:
                msg += "It might be possible to solve this slot collision\n"
            else:
                msg += "It might be possible to solve these slot collisions\n"
            msg += "by applying one of the following solutions:\n"

        def print_change(change, indent=""):
            mymsg = ""
            for pkg in change:
                changes = []
                for flag, state in change[pkg].items():
                    if state:
                        changes.append(colorize("red", "+" + flag))
                    else:
                        changes.append(colorize("blue", "-" + flag))
                mymsg += (
                    indent
                    + "- "
                    + pkg.cpv
                    + " (Change USE: %s" % " ".join(changes)
                    + ")\n"
                )
            mymsg += "\n"
            return mymsg

        if len(self.changes) == 1:
            msg += print_change(self.changes[0], "   ")
        else:
            for change in self.changes:
                msg += "  Solution: Apply all of:\n"
                msg += print_change(change, "     ")

        return msg

    def _check_configuration(
        self, config, all_conflict_atoms_by_slotatom, conflict_nodes
    ):
        """
        Given a configuration, required use changes are computed and checked to
        make sure that no new conflict is introduced. Returns a solution or None.
        """
        _pkg_use_enabled = self.depgraph._pkg_use_enabled
        # An installed package can only be part of a valid configuration if it has no
        # pending use changes. Otherwise the ebuild will be pulled in again.
        for pkg in config:
            if not pkg.installed:
                continue

            for root, atom, pkgs in self.all_conflicts:
                if pkg not in pkgs:
                    continue
                for other_pkg in pkgs:
                    if other_pkg == pkg:
                        continue
                    if pkg.iuse.all.symmetric_difference(
                        other_pkg.iuse.all
                    ) or _pkg_use_enabled(pkg).symmetric_difference(
                        _pkg_use_enabled(other_pkg)
                    ):
                        if self.debug:
                            writemsg(
                                (
                                    "%s has pending USE changes. "
                                    "Rejecting configuration.\n"
                                )
                                % (pkg,),
                                noiselevel=-1,
                            )
                        return False

        # A list of dicts. Keeps one dict per slot conflict. [ { flag1: "enabled" }, { flag2: "disabled" } ]
        all_involved_flags = []

        # Go through all slot conflicts
        for idx, pkg in enumerate(config):
            involved_flags = {}
            for ppkg, atom in all_conflict_atoms_by_slotatom[idx]:
                if not atom.package:
                    continue

                if ppkg in conflict_nodes and not ppkg in config:
                    # The parent is part of a slot conflict itself and is
                    # not part of the current config.
                    continue

                i = InternalPackageSet(initial_atoms=(atom,))
                if i.findAtomForPackage(pkg, modified_use=_pkg_use_enabled(pkg)):
                    continue

                i = InternalPackageSet(initial_atoms=(atom.without_use,))
                if not i.findAtomForPackage(pkg, modified_use=_pkg_use_enabled(pkg)):
                    # Version range does not match.
                    if self.debug:
                        writemsg(
                            (
                                "%s does not satify all version "
                                "requirements. Rejecting configuration.\n"
                            )
                            % (pkg,),
                            noiselevel=-1,
                        )
                    return False

                if not pkg.iuse.is_valid_flag(atom.unevaluated_atom.use.required):
                    # Missing IUSE.
                    # FIXME: This needs to support use dep defaults.
                    if self.debug:
                        writemsg(
                            (
                                "%s misses needed flags from IUSE."
                                " Rejecting configuration.\n"
                            )
                            % (pkg,),
                            noiselevel=-1,
                        )
                    return False

                if not isinstance(ppkg, Package) or ppkg.installed:
                    # We cannot assume that it's possible to reinstall the package. Do not
                    # check if some of its atom has use.conditional
                    violated_atom = atom.violated_conditionals(
                        _pkg_use_enabled(pkg), pkg.iuse.is_valid_flag
                    )
                else:
                    violated_atom = atom.unevaluated_atom.violated_conditionals(
                        _pkg_use_enabled(pkg),
                        pkg.iuse.is_valid_flag,
                        parent_use=_pkg_use_enabled(ppkg),
                    )
                    if violated_atom.use is None:
                        # It's possible for autounmask to change
                        # parent_use such that the unevaluated form
                        # of the atom now matches, even though the
                        # earlier evaluated form (from before
                        # autounmask changed parent_use) does not.
                        # In this case (see bug #374423), it's
                        # expected that violated_atom.use is None.
                        continue

                if pkg.installed and (
                    violated_atom.use.enabled or violated_atom.use.disabled
                ):
                    # We can't change USE of an installed package (only of an ebuild, but that is already
                    # part of the conflict, isn't it?)
                    if self.debug:
                        writemsg(
                            (
                                "%s: installed package would need USE"
                                " changes. Rejecting configuration.\n"
                            )
                            % (pkg,),
                            noiselevel=-1,
                        )
                    return False

                # Compute the required USE changes. A flag can be forced to "enabled" or "disabled",
                # it can be in the conditional state "cond" that allows both values or in the
                # "contradiction" state, which means that some atoms insist on different values
                # for this flag and those kill this configuration.
                for flag in violated_atom.use.required:
                    state = involved_flags.get(flag, "")

                    if flag in violated_atom.use.enabled:
                        if state in ("", "cond", "enabled"):
                            state = "enabled"
                        else:
                            state = "contradiction"
                    elif flag in violated_atom.use.disabled:
                        if state in ("", "cond", "disabled"):
                            state = "disabled"
                        else:
                            state = "contradiction"
                    else:
                        if state == "":
                            state = "cond"

                    involved_flags[flag] = state

            if pkg.installed:
                # We don't change the installed pkg's USE. Force all involved flags
                # to the same value as the installed package has it.
                for flag in involved_flags:
                    if involved_flags[flag] == "enabled":
                        if not flag in _pkg_use_enabled(pkg):
                            involved_flags[flag] = "contradiction"
                    elif involved_flags[flag] == "disabled":
                        if flag in _pkg_use_enabled(pkg):
                            involved_flags[flag] = "contradiction"
                    elif involved_flags[flag] == "cond":
                        if flag in _pkg_use_enabled(pkg):
                            involved_flags[flag] = "enabled"
                        else:
                            involved_flags[flag] = "disabled"

            for flag, state in involved_flags.items():
                if state == "contradiction":
                    if self.debug:
                        writemsg(
                            "Contradicting requirements found for flag "
                            + flag
                            + ". Rejecting configuration.\n",
                            noiselevel=-1,
                        )
                    return False

            all_involved_flags.append(involved_flags)

        if self.debug:
            writemsg("All involved flags:\n", noiselevel=-1)
            for idx, involved_flags in enumerate(all_involved_flags):
                writemsg("   {}\n".format(config[idx]), noiselevel=-1)
                for flag, state in involved_flags.items():
                    writemsg("     " + flag + ": " + state + "\n", noiselevel=-1)

        solutions = []
        sol_gen = _solution_candidate_generator(all_involved_flags)
        checked = 0
        while True:
            candidate = sol_gen.get_candidate()
            if not candidate:
                break
            solution = self._check_solution(
                config, candidate, all_conflict_atoms_by_slotatom
            )
            checked += 1
            if solution:
                solutions.append(solution)

            if checked >= self._check_configuration_max:
                # TODO: Implement early elimination for candidates that would
                # change forced or masked flags, and don't count them here.
                if self.debug:
                    writemsg(
                        "\nAborting _check_configuration due to "
                        "excessive number of candidates.\n",
                        noiselevel=-1,
                    )
                break

        if self.debug:
            if not solutions:
                writemsg(
                    "No viable solutions. Rejecting configuration.\n", noiselevel=-1
                )
        return solutions

    def _force_flag_for_package(self, required_changes, pkg, flag, state):
        """
        Adds a USE change to required_changes. Sets the target state to
        "contradiction" if a flag is forced to conflicting values.
        """
        _pkg_use_enabled = self.depgraph._pkg_use_enabled

        if state == "disabled":
            changes = required_changes.get(pkg, {})
            flag_change = changes.get(flag, "")
            if flag_change == "enabled":
                flag_change = "contradiction"
            elif flag in _pkg_use_enabled(pkg):
                flag_change = "disabled"

            changes[flag] = flag_change
            required_changes[pkg] = changes
        elif state == "enabled":
            changes = required_changes.get(pkg, {})
            flag_change = changes.get(flag, "")
            if flag_change == "disabled":
                flag_change = "contradiction"
            else:
                flag_change = "enabled"

            changes[flag] = flag_change
            required_changes[pkg] = changes

    def _check_solution(
        self, config, all_involved_flags, all_conflict_atoms_by_slotatom
    ):
        """
        Given a configuration and all involved flags, all possible settings for the involved
        flags are checked if they solve the slot conflict.
        """
        _pkg_use_enabled = self.depgraph._pkg_use_enabled

        if self.debug:
            # The code is a bit verbose, because the states might not
            # be a string, but a _value_helper.
            msg = "Solution candidate: "
            msg += "["
            first = True
            for involved_flags in all_involved_flags:
                if first:
                    first = False
                else:
                    msg += ", "
                msg += "{"
                inner_first = True
                for flag, state in involved_flags.items():
                    if inner_first:
                        inner_first = False
                    else:
                        msg += ", "
                    msg += flag + ": {}".format(state)
                msg += "}"
            msg += "]\n"
            writemsg(msg, noiselevel=-1)

        required_changes = {}
        for idx, pkg in enumerate(config):
            if not pkg.installed:
                # We can't change the USE of installed packages.
                for flag in all_involved_flags[idx]:
                    if not pkg.iuse.is_valid_flag(flag):
                        continue
                    state = all_involved_flags[idx][flag]
                    self._force_flag_for_package(required_changes, pkg, flag, state)

            # Go through all (parent, atom) pairs for the current slot conflict.
            for ppkg, atom in all_conflict_atoms_by_slotatom[idx]:
                if not atom.package:
                    continue
                use = atom.unevaluated_atom.use
                if not use:
                    # No need to force something for an atom without USE conditionals.
                    # These atoms are already satisfied.
                    continue
                for flag in all_involved_flags[idx]:
                    state = all_involved_flags[idx][flag]

                    if flag not in use.required or not use.conditional:
                        continue
                    if flag in use.conditional.enabled:
                        # [flag?]
                        if state == "enabled":
                            # no need to change anything, the atom won't
                            # force -flag on pkg
                            pass
                        elif state == "disabled":
                            # if flag is enabled we get [flag] -> it must be disabled
                            self._force_flag_for_package(
                                required_changes, ppkg, flag, "disabled"
                            )
                    elif flag in use.conditional.disabled:
                        # [!flag?]
                        if state == "enabled":
                            # if flag is enabled we get [-flag] -> it must be disabled
                            self._force_flag_for_package(
                                required_changes, ppkg, flag, "disabled"
                            )
                        elif state == "disabled":
                            # no need to change anything, the atom won't
                            # force +flag on pkg
                            pass
                    elif flag in use.conditional.equal:
                        # [flag=]
                        if state == "enabled":
                            # if flag is disabled we get [-flag] -> it must be enabled
                            self._force_flag_for_package(
                                required_changes, ppkg, flag, "enabled"
                            )
                        elif state == "disabled":
                            # if flag is enabled we get [flag] -> it must be disabled
                            self._force_flag_for_package(
                                required_changes, ppkg, flag, "disabled"
                            )
                    elif flag in use.conditional.not_equal:
                        # [!flag=]
                        if state == "enabled":
                            # if flag is enabled we get [-flag] -> it must be disabled
                            self._force_flag_for_package(
                                required_changes, ppkg, flag, "disabled"
                            )
                        elif state == "disabled":
                            # if flag is disabled we get [flag] -> it must be enabled
                            self._force_flag_for_package(
                                required_changes, ppkg, flag, "enabled"
                            )

        is_valid_solution = True
        for pkg in required_changes:
            for state in required_changes[pkg].values():
                if not state in ("enabled", "disabled"):
                    is_valid_solution = False

        if not is_valid_solution:
            return None

        # Check if all atoms are satisfied after the changes are applied.
        for idx, pkg in enumerate(config):
            new_use = _pkg_use_enabled(pkg)
            if pkg in required_changes:
                old_use = pkg.use.enabled
                new_use = set(new_use)
                for flag, state in required_changes[pkg].items():
                    if state == "enabled":
                        new_use.add(flag)
                    elif state == "disabled":
                        new_use.discard(flag)
                if not new_use.symmetric_difference(old_use):
                    # avoid copying the package in findAtomForPackage if possible
                    new_use = old_use

            for ppkg, atom in all_conflict_atoms_by_slotatom[idx]:
                if not atom.package:
                    continue
                if not hasattr(ppkg, "use"):
                    # It's a SetArg or something like that.
                    continue
                ppkg_new_use = set(_pkg_use_enabled(ppkg))
                if ppkg in required_changes:
                    for flag, state in required_changes[ppkg].items():
                        if state == "enabled":
                            ppkg_new_use.add(flag)
                        elif state == "disabled":
                            ppkg_new_use.discard(flag)

                new_atom = atom.unevaluated_atom.evaluate_conditionals(ppkg_new_use)
                i = InternalPackageSet(initial_atoms=(new_atom,))
                if not i.findAtomForPackage(pkg, new_use):
                    # We managed to create a new problem with our changes.
                    is_valid_solution = False
                    if self.debug:
                        writemsg(
                            (
                                "new conflict introduced: %s"
                                " does not match %s from %s\n"
                            )
                            % (pkg, new_atom, ppkg),
                            noiselevel=-1,
                        )
                    break

            if not is_valid_solution:
                break

        # Make sure the changes don't violate REQUIRED_USE
        for pkg in required_changes:
            required_use = pkg._metadata.get("REQUIRED_USE")
            if not required_use:
                continue

            use = set(_pkg_use_enabled(pkg))
            for flag, state in required_changes[pkg].items():
                if state == "enabled":
                    use.add(flag)
                else:
                    use.discard(flag)

            if not check_required_use(required_use, use, pkg.iuse.is_valid_flag):
                is_valid_solution = False
                break

        if is_valid_solution and required_changes:
            return required_changes
        return None


class _configuration_generator:
    def __init__(self, conflict_pkgs):
        # reorder packages such that installed packages come last
        self.conflict_pkgs = []
        for pkgs in conflict_pkgs:
            new_pkgs = []
            for pkg in pkgs:
                if not pkg.installed:
                    new_pkgs.append(pkg)
            for pkg in pkgs:
                if pkg.installed:
                    new_pkgs.append(pkg)
            self.conflict_pkgs.append(new_pkgs)

        self.solution_ids = []
        for pkgs in self.conflict_pkgs:
            self.solution_ids.append(0)
        self._is_first_solution = True

    def get_configuration(self):
        if self._is_first_solution:
            self._is_first_solution = False
        else:
            if not self._next():
                return None

        solution = []
        for idx, pkgs in enumerate(self.conflict_pkgs):
            solution.append(pkgs[self.solution_ids[idx]])
        return solution

    def _next(self, id=None):  # pylint: disable=redefined-builtin
        solution_ids = self.solution_ids
        conflict_pkgs = self.conflict_pkgs

        if id is None:
            id = len(solution_ids) - 1

        if solution_ids[id] == len(conflict_pkgs[id]) - 1:
            if id > 0:
                return self._next(id=id - 1)
            return False

        solution_ids[id] += 1
        for other_id in range(id + 1, len(solution_ids)):
            solution_ids[other_id] = 0
        return True


class _solution_candidate_generator:
    class _value_helper:
        def __init__(self, value=None):
            self.value = value

        def __eq__(self, other):
            if isinstance(other, str):
                return self.value == other
            return self.value == other.value

        def __str__(self):
            return "{}".format(self.value)

    def __init__(self, all_involved_flags):
        # A copy of all_involved_flags with all "cond" values
        # replaced by a _value_helper object.
        self.all_involved_flags = []

        # A list tracking references to all used _value_helper
        # objects.
        self.conditional_values = []

        for involved_flags in all_involved_flags:
            new_involved_flags = {}
            for flag, state in involved_flags.items():
                if state in ("enabled", "disabled"):
                    new_involved_flags[flag] = state
                else:
                    v = self._value_helper("disabled")
                    new_involved_flags[flag] = v
                    self.conditional_values.append(v)
            self.all_involved_flags.append(new_involved_flags)

        self._is_first_solution = True

    def get_candidate(self):
        if self._is_first_solution:
            self._is_first_solution = False
        else:
            if not self._next():
                return None

        return self.all_involved_flags

    def _next(self, id=None):  # pylint: disable=redefined-builtin
        values = self.conditional_values

        if not values:
            return False

        if id is None:
            id = len(values) - 1

        if values[id].value == "enabled":
            if id > 0:
                return self._next(id=id - 1)
            return False

        values[id].value = "enabled"
        for other_id in range(id + 1, len(values)):
            values[other_id].value = "disabled"
        return True



# test_uniqueArray.py -- Portage Unit Testing Functionality
# Copyright 2006-2011 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage import os
from portage.tests import TestCase
from portage.util import unique_array


class UniqueArrayTestCase(TestCase):
    def testUniqueArrayPass(self):
        """
        test portage.util.uniqueArray()
        """

        tests = [
            (["a", "a", "a", os, os, [], [], []], ["a", os, []]),
            ([1, 1, 1, 2, 3, 4, 4], [1, 2, 3, 4]),
        ]

        for test in tests:
            result = unique_array(test[0])
            for item in test[1]:
                number = result.count(item)
                self.assertFalse(
                    number != 1,
                    msg=("%s contains %s of %s, " "should be only 1")
                    % (result, number, item),
                )



# Copright Gentoo Foundation 2006-2020
# Portage Unit Testing Functionality

import tempfile
from os import urandom

from portage import os
from portage import shutil
from portage.util._compare_files import compare_files
from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import ResolverPlayground
from portage.gpkg import gpkg


class test_gpkg_metadata_case(TestCase):
    def test_gpkg_update_metadata(self):
        playground = ResolverPlayground(
            user_config={
                "make.conf": ('BINPKG_COMPRESS="gzip"',),
            }
        )
        tmpdir = tempfile.mkdtemp()

        try:
            settings = playground.settings
            orig_full_path = os.path.join(tmpdir, "orig/")
            os.makedirs(orig_full_path)
            with open(os.path.join(orig_full_path, "test"), "wb") as test_file:
                test_file.write(urandom(1048576))

            gpkg_file_loc = os.path.join(tmpdir, "test.gpkg.tar")
            test_gpkg = gpkg(settings, "test", gpkg_file_loc)

            meta = {"test1": b"1234567890", "test2": b"abcdef"}

            test_gpkg.compress(os.path.join(tmpdir, "orig"), meta)

            meta_result = test_gpkg.get_metadata()
            self.assertEqual(meta, meta_result)

            meta_new = {"test3": b"0987654321", "test4": b"XXXXXXXX"}
            test_gpkg.update_metadata(meta_new)

            meta_result = test_gpkg.get_metadata()
            self.assertEqual(meta_new, meta_result)

            test_gpkg.decompress(os.path.join(tmpdir, "test"))
            r = compare_files(
                os.path.join(tmpdir, "orig/" + "test"),
                os.path.join(tmpdir, "test/" + "test"),
                skipped_types=("atime", "mtime", "ctime"),
            )
            self.assertEqual(r, ())
        finally:
            shutil.rmtree(tmpdir)
            playground.cleanup()



# Copyright 2010-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

__all__ = (
    "case_insensitive_vars",
    "default_globals",
    "env_blacklist",
    "environ_filter",
    "environ_whitelist",
    "environ_whitelist_re",
)

import re

# Blacklisted variables are internal variables that are never allowed
# to enter the config instance from the external environment or
# configuration files.
env_blacklist = frozenset(
    (
        "A",
        "AA",
        "BASH_FUNC____in_portage_iuse%%",
        "BDEPEND",
        "BROOT",
        "CATEGORY",
        "DEPEND",
        "DESCRIPTION",
        "DOCS",
        "EAPI",
        "EBUILD_FORCE_TEST",
        "EBUILD_PHASE",
        "EBUILD_PHASE_FUNC",
        "EBUILD_SKIP_MANIFEST",
        "ED",
        "EMERGE_FROM",
        "EPREFIX",
        "EROOT",
        "GREP_OPTIONS",
        "HOMEPAGE",
        "IDEPEND",
        "INHERITED",
        "IUSE",
        "IUSE_EFFECTIVE",
        "KEYWORDS",
        "LICENSE",
        "MERGE_TYPE",
        "PDEPEND",
        "PF",
        "PKGUSE",
        "PORTAGE_BACKGROUND",
        "PORTAGE_BACKGROUND_UNMERGE",
        "PORTAGE_BUILDDIR_LOCKED",
        "PORTAGE_BUILT_USE",
        "PORTAGE_CONFIGROOT",
        "PORTAGE_EXPLICIT_INHERIT",
        "PORTAGE_INTERNAL_CALLER",
        "PORTAGE_IUSE",
        "PORTAGE_NONFATAL",
        "PORTAGE_PIPE_FD",
        "PORTAGE_REPO_NAME",
        "PORTAGE_USE",
        "PROPERTIES",
        "RDEPEND",
        "REPOSITORY",
        "REQUIRED_USE",
        "RESTRICT",
        "ROOT",
        "SANDBOX_LOG",
        "SLOT",
        "SRC_URI",
        "_",
    )
)

# Whitelisted variables are always allowed to enter the ebuild
# environment. Generally, this only includes special portage
# variables. Ebuilds can unset variables that are not whitelisted
# and rely on them remaining unset for future phases, without them
# leaking back in from various locations (bug #189417). It's very
# important to set our special BASH_ENV variable in the ebuild
# environment in order to prevent sandbox from sourcing /etc/profile
# in it's bashrc (causing major leakage).
environ_whitelist = frozenset(
    (
        "A",
        "AA",
        "ACCEPT_LICENSE",
        "BASH_ENV",
        "BASH_FUNC____in_portage_iuse%%",
        "BINPKG_FORMAT",
        "BROOT",
        "BUILD_ID",
        "BUILD_PREFIX",
        "CATEGORY",
        "COLUMNS",
        "D",
        "DISTDIR",
        "DOC_SYMLINKS_DIR",
        "EAPI",
        "EBUILD",
        "EBUILD_FORCE_TEST",
        "EBUILD_PHASE",
        "EBUILD_PHASE_FUNC",
        "ECLASSDIR",
        "ECLASS_DEPTH",
        "ED",
        "EMERGE_FROM",
        "ENV_UNSET",
        "EPREFIX",
        "EROOT",
        "ESYSROOT",
        "FEATURES",
        "FILESDIR",
        "HOME",
        "MERGE_TYPE",
        "NOCOLOR",
        "P",
        "PATH",
        "PF",
        "PKGDIR",
        "PKGUSE",
        "PKG_LOGDIR",
        "PKG_TMPDIR",
        "PM_EBUILD_HOOK_DIR",
        "PN",
        "PORTAGE_ACTUAL_DISTDIR",
        "PORTAGE_ARCHLIST",
        "PORTAGE_BASHRC_FILES",
        "PORTAGE_BASHRC",
        "PORTAGE_BINPKG_FILE",
        "PORTAGE_BINPKG_TAR_OPTS",
        "PORTAGE_BINPKG_TMPFILE",
        "PORTAGE_BIN_PATH",
        "PORTAGE_BUILDDIR",
        "PORTAGE_BUILD_GROUP",
        "PORTAGE_BUILD_USER",
        "PORTAGE_BUNZIP2_COMMAND",
        "PORTAGE_BZIP2_COMMAND",
        "PORTAGE_COLORMAP",
        "PORTAGE_COMPRESS",
        "PORTAGE_COMPRESSION_COMMAND",
        "PORTAGE_COMPRESS_EXCLUDE_SUFFIXES",
        "PORTAGE_CONFIGROOT",
        "PORTAGE_DEBUG",
        "PORTAGE_DEPCACHEDIR",
        "PORTAGE_DOHTML_UNWARNED_SKIPPED_EXTENSIONS",
        "PORTAGE_DOHTML_UNWARNED_SKIPPED_FILES",
        "PORTAGE_DOHTML_WARN_ON_SKIPPED_FILES",
        "PORTAGE_EBUILD_EXIT_FILE",
        "PORTAGE_FEATURES",
        "PORTAGE_GID",
        "PORTAGE_GRPNAME",
        "PORTAGE_INTERNAL_CALLER",
        "PORTAGE_INST_GID",
        "PORTAGE_INST_UID",
        "PORTAGE_IPC_DAEMON",
        "PORTAGE_IUSE",
        "PORTAGE_ECLASS_LOCATIONS",
        "PORTAGE_LOG_FILE",
        "PORTAGE_OVERRIDE_EPREFIX",
        "PORTAGE_PIPE_FD",
        "PORTAGE_PROPERTIES",
        "PORTAGE_PYM_PATH",
        "PORTAGE_PYTHON",
        "PORTAGE_PYTHONPATH",
        "PORTAGE_QUIET",
        "PORTAGE_REPO_NAME",
        "PORTAGE_REPOSITORIES",
        "PORTAGE_RESTRICT",
        "PORTAGE_SIGPIPE_STATUS",
        "PORTAGE_SOCKS5_PROXY",
        "PORTAGE_TMPDIR",
        "PORTAGE_UPDATE_ENV",
        "PORTAGE_USERNAME",
        "PORTAGE_VERBOSE",
        "PORTAGE_WORKDIR_MODE",
        "PORTAGE_XATTR_EXCLUDE",
        "PORTDIR",
        "PORTDIR_OVERLAY",
        "PR",
        "PREROOTPATH",
        "PV",
        "PVR",
        "PYTHONDONTWRITEBYTECODE",
        "REPLACING_VERSIONS",
        "REPLACED_BY_VERSION",
        "ROOT",
        "ROOTPATH",
        "SANDBOX_LOG",
        "SYSROOT",
        "T",
        "TMP",
        "TMPDIR",
        "USE_EXPAND",
        "USE_ORDER",
        "WORKDIR",
        "XARGS",
        "__PORTAGE_TEST_HARDLINK_LOCKS",
        # user config variables
        "DOC_SYMLINKS_DIR",
        "INSTALL_MASK",
        "PKG_INSTALL_MASK",
        # misc variables inherited from the calling environment
        "COLORTERM",
        "DISPLAY",
        "EDITOR",
        "LESS",
        "LESSOPEN",
        "LOGNAME",
        "LS_COLORS",
        "PAGER",
        "TERM",
        "TERMCAP",
        "USER",
        "ftp_proxy",
        "http_proxy",
        "no_proxy",
        # tempdir settings
        "TMPDIR",
        "TEMP",
        "TMP",
        # localization settings
        "LANG",
        "LC_COLLATE",
        "LC_CTYPE",
        "LC_MESSAGES",
        "LC_MONETARY",
        "LC_NUMERIC",
        "LC_TIME",
        "LC_PAPER",
        "LC_ALL",
        # other variables inherited from the calling environment
        "CVS_RSH",
        "ECHANGELOG_USER",
        "GPG_AGENT_INFO",
        "SSH_AGENT_PID",
        "SSH_AUTH_SOCK",
        "STY",
        "WINDOW",
        "XAUTHORITY",
    )
)

environ_whitelist_re = re.compile(r"^(CCACHE_|DISTCC_).*")

# Filter selected variables in the config.environ() method so that
# they don't needlessly propagate down into the ebuild environment.
# Exclude anything that could be extremely long here (like SRC_URI)
# since that could cause execve() calls to fail with E2BIG errors. For
# example, see bug #262647.
environ_filter = frozenset(
    (
        "DEPEND",
        "RDEPEND",
        "PDEPEND",
        "SRC_URI",
        "BDEPEND",
        "IDEPEND",
        # misc variables inherited from the calling environment
        "INFOPATH",
        "MANPATH",
        "USER",
        # variables that break bash
        "HISTFILE",
        "POSIXLY_CORRECT",
        # portage config variables and variables set directly by portage
        "ACCEPT_CHOSTS",
        "ACCEPT_KEYWORDS",
        "ACCEPT_PROPERTIES",
        "ACCEPT_RESTRICT",
        "AUTOCLEAN",
        "BINPKG_COMPRESS",
        "BINPKG_COMPRESS_FLAGS",
        "CLEAN_DELAY",
        "COLLISION_IGNORE",
        "CONFIG_PROTECT",
        "CONFIG_PROTECT_MASK",
        "EGENCACHE_DEFAULT_OPTS",
        "EMERGE_DEFAULT_OPTS",
        "EMERGE_LOG_DIR",
        "EMERGE_WARNING_DELAY",
        "FETCHCOMMAND",
        "FETCHCOMMAND_FTP",
        "FETCHCOMMAND_HTTP",
        "FETCHCOMMAND_HTTPS",
        "FETCHCOMMAND_RSYNC",
        "FETCHCOMMAND_SFTP",
        "GENTOO_MIRRORS",
        "NOCONFMEM",
        "O",
        "PORTAGE_BACKGROUND",
        "PORTAGE_BACKGROUND_UNMERGE",
        "PORTAGE_BINHOST",
        "PORTAGE_BINPKG_FORMAT",
        "PORTAGE_BUILDDIR_LOCKED",
        "PORTAGE_CHECKSUM_FILTER",
        "PORTAGE_ELOG_CLASSES",
        "PORTAGE_ELOG_MAILFROM",
        "PORTAGE_ELOG_MAILSUBJECT",
        "PORTAGE_ELOG_MAILURI",
        "PORTAGE_ELOG_SYSTEM",
        "PORTAGE_FETCH_CHECKSUM_TRY_MIRRORS",
        "PORTAGE_FETCH_RESUME_MIN_SIZE",
        "PORTAGE_GPG_DIR",
        "PORTAGE_GPG_KEY",
        "PORTAGE_GPG_SIGNING_COMMAND",
        "PORTAGE_IONICE_COMMAND",
        "PORTAGE_PACKAGE_EMPTY_ABORT",
        "PORTAGE_REPO_DUPLICATE_WARN",
        "PORTAGE_RO_DISTDIRS",
        "PORTAGE_RSYNC_EXTRA_OPTS",
        "PORTAGE_RSYNC_OPTS",
        "PORTAGE_RSYNC_RETRIES",
        "PORTAGE_SSH_OPTS",
        "PORTAGE_SYNC_STALE",
        "PORTAGE_USE",
        "PORTAGE_LOG_FILTER_FILE_CMD",
        "PORTAGE_LOGDIR",
        "PORTAGE_LOGDIR_CLEAN",
        "QUICKPKG_DEFAULT_OPTS",
        "REPOMAN_DEFAULT_OPTS",
        "RESUMECOMMAND",
        "RESUMECOMMAND_FTP",
        "RESUMECOMMAND_HTTP",
        "RESUMECOMMAND_HTTPS",
        "RESUMECOMMAND_RSYNC",
        "RESUMECOMMAND_SFTP",
        "UNINSTALL_IGNORE",
        "USE_EXPAND_HIDDEN",
        "USE_ORDER",
        "__PORTAGE_HELPER",
        # No longer supported variables
        "SYNC",
    )
)

# Variables that are not allowed to have per-repo or per-package
# settings.
global_only_vars = frozenset(
    [
        "CONFIG_PROTECT",
    ]
)

default_globals = {
    "ACCEPT_PROPERTIES": "*",
    "PORTAGE_BZIP2_COMMAND": "bzip2",
}

validate_commands = (
    "PORTAGE_BZIP2_COMMAND",
    "PORTAGE_BUNZIP2_COMMAND",
    "PORTAGE_LOG_FILTER_FILE_CMD",
)

# To enhance usability, make some vars case insensitive
# by forcing them to lower case.
case_insensitive_vars = (
    "AUTOCLEAN",
    "NOCOLOR",
)



# Copyright 2018-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

__all__ = (
    "RetryError",
    "retry",
)

import functools

from portage.exception import PortageException
from portage.util.futures import asyncio


class RetryError(PortageException):
    """Raised when retry fails."""

    def __init__(self):
        PortageException.__init__(self, "retry error")


def retry(
    try_max=None,
    try_timeout=None,
    overall_timeout=None,
    delay_func=None,
    reraise=False,
    loop=None,
):
    """
    Create and return a retry decorator. The decorator is intended to
    operate only on a coroutine function.

    @param try_max: maximum number of tries
    @type try_max: int or None
    @param try_timeout: number of seconds to wait for a try to succeed
            before cancelling it, which is only effective if func returns
            tasks that support cancellation
    @type try_timeout: float or None
    @param overall_timeout: number of seconds to wait for retires to
            succeed before aborting, which is only effective if func returns
            tasks that support cancellation
    @type overall_timeout: float or None
    @param delay_func: function that takes an int argument corresponding
            to the number of previous tries and returns a number of seconds
            to wait before the next try
    @type delay_func: callable
    @param reraise: Reraise the last exception, instead of RetryError
    @type reraise: bool
    @param loop: event loop
    @type loop: EventLoop
    @return: func decorated with retry support
    @rtype: callable
    """
    return functools.partial(
        _retry_wrapper, loop, try_max, try_timeout, overall_timeout, delay_func, reraise
    )


def _retry_wrapper(
    _loop, try_max, try_timeout, overall_timeout, delay_func, reraise, func, loop=None
):
    """
    Create and return a decorated function.
    """
    return functools.partial(
        _retry,
        loop or _loop,
        try_max,
        try_timeout,
        overall_timeout,
        delay_func,
        reraise,
        func,
    )


def _retry(
    loop,
    try_max,
    try_timeout,
    overall_timeout,
    delay_func,
    reraise,
    func,
    *args,
    **kwargs
):
    """
    Retry coroutine, used to implement retry decorator.

    @return: func return value
    @rtype: asyncio.Future (or compatible)
    """
    loop = asyncio._wrap_loop(loop)
    future = loop.create_future()
    _Retry(
        future,
        loop,
        try_max,
        try_timeout,
        overall_timeout,
        delay_func,
        reraise,
        functools.partial(func, *args, **kwargs),
    )
    return future


class _Retry:
    def __init__(
        self,
        future,
        loop,
        try_max,
        try_timeout,
        overall_timeout,
        delay_func,
        reraise,
        func,
    ):
        self._future = future
        self._loop = loop
        self._try_max = try_max
        self._try_timeout = try_timeout
        self._delay_func = delay_func
        self._reraise = reraise
        self._func = func

        self._try_timeout_handle = None
        self._overall_timeout_handle = None
        self._overall_timeout_expired = None
        self._tries = 0
        self._current_task = None
        self._previous_result = None

        future.add_done_callback(self._cancel_callback)
        if overall_timeout is not None:
            self._overall_timeout_handle = loop.call_later(
                overall_timeout, self._overall_timeout_callback
            )
        self._begin_try()

    def _cancel_callback(self, future):
        if future.cancelled() and self._current_task is not None:
            self._current_task.cancel()

    def _try_timeout_callback(self):
        self._try_timeout_handle = None
        self._current_task.cancel()

    def _overall_timeout_callback(self):
        self._overall_timeout_handle = None
        self._overall_timeout_expired = True
        self._current_task.cancel()
        self._retry_error()

    def _begin_try(self):
        self._tries += 1
        self._current_task = asyncio.ensure_future(self._func(), loop=self._loop)
        self._current_task.add_done_callback(self._try_done)
        if self._try_timeout is not None:
            self._try_timeout_handle = self._loop.call_later(
                self._try_timeout, self._try_timeout_callback
            )

    def _try_done(self, future):
        self._current_task = None

        if self._try_timeout_handle is not None:
            self._try_timeout_handle.cancel()
            self._try_timeout_handle = None

        if not future.cancelled():
            # consume exception, so that the event loop
            # exception handler does not report it
            future.exception()

        if self._overall_timeout_expired:
            return

        try:
            if self._future.cancelled():
                return

            self._previous_result = future
            if not (future.cancelled() or future.exception() is not None):
                # success
                self._future.set_result(future.result())
                return
        finally:
            if self._future.done() and self._overall_timeout_handle is not None:
                self._overall_timeout_handle.cancel()
                self._overall_timeout_handle = None

        if self._try_max is not None and self._tries >= self._try_max:
            self._retry_error()
            return

        if self._delay_func is not None:
            delay = self._delay_func(self._tries)
            self._current_task = self._loop.call_later(delay, self._delay_done)
            return

        self._begin_try()

    def _delay_done(self):
        self._current_task = None

        if self._future.cancelled() or self._overall_timeout_expired:
            return

        self._begin_try()

    def _retry_error(self):
        if self._previous_result is None or self._previous_result.cancelled():
            cause = asyncio.TimeoutError()
        else:
            cause = self._previous_result.exception()

        if self._reraise:
            e = cause
        else:
            e = RetryError()
            e.__cause__ = cause

        self._future.set_exception(e)



# Copyright 1999-2013 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import logging

from _emerge.SpawnProcess import SpawnProcess
import portage
from portage.localization import _
from portage.util.compression_probe import (
    compression_probe,
    _compressors,
)
from portage.util.cpuinfo import makeopts_to_job_count
from portage.process import find_binary
from portage.util import (
    shlex_split,
    varexpand,
)
from portage.exception import InvalidBinaryPackageFormat
from portage.binpkg import get_binpkg_format
import signal
import subprocess
import tarfile


class BinpkgExtractorAsync(SpawnProcess):

    __slots__ = ("features", "image_dir", "pkg", "pkg_path")

    _shell_binary = portage.const.BASH_BINARY

    def _start(self):
        binpkg_format = get_binpkg_format(self.pkg_path)
        if binpkg_format == "xpak":
            self._xpak_start()
        else:
            raise InvalidBinaryPackageFormat(
                f"{self.pkg_path} is not a valid xpak binary package"
            )

    def _xpak_start(self):
        tar_options = ""
        if "xattr" in self.features:
            process = subprocess.Popen(
                ["tar", "--help"], stdout=subprocess.PIPE, stderr=subprocess.PIPE
            )
            output = process.communicate()[0]
            if b"--xattrs" in output:
                tar_options = ["--xattrs", "--xattrs-include='*'"]
                for x in portage.util.shlex_split(
                    self.env.get("PORTAGE_XATTR_EXCLUDE", "")
                ):
                    tar_options.append(portage._shell_quote("--xattrs-exclude=%s" % x))
                tar_options = " ".join(tar_options)

        decomp = _compressors.get(compression_probe(self.pkg_path))
        if decomp is not None:
            decomp_cmd = decomp.get("decompress")
            decomp_cmd = decomp_cmd.replace(
                "{JOBS}", str(makeopts_to_job_count(self.env.get("MAKEOPTS", "1")))
            )
        elif tarfile.is_tarfile(
            portage._unicode_encode(
                self.pkg_path, encoding=portage._encodings["fs"], errors="strict"
            )
        ):
            decomp_cmd = "cat"
            decomp = {
                "compress": "cat",
                "package": "sys-apps/coreutils",
            }
        else:
            decomp_cmd = None
        if decomp_cmd is None:
            self.scheduler.output(
                "!!! %s\n"
                % _("File compression header unrecognized: %s")
                % self.pkg_path,
                log_path=self.logfile,
                background=self.background,
                level=logging.ERROR,
            )
            self.returncode = 1
            self._async_wait()
            return

        try:
            decompression_binary = shlex_split(varexpand(decomp_cmd, mydict=self.env))[
                0
            ]
        except IndexError:
            decompression_binary = ""

        if find_binary(decompression_binary) is None:
            # Try alternative command if it exists
            if decomp.get("decompress_alt"):
                decomp_cmd = decomp.get("decompress_alt")
            try:
                decompression_binary = shlex_split(
                    varexpand(decomp_cmd, mydict=self.env)
                )[0]
            except IndexError:
                decompression_binary = ""

            if find_binary(decompression_binary) is None:
                missing_package = decomp.get("package")
                self.scheduler.output(
                    "!!! %s\n"
                    % _(
                        "File compression unsupported %s.\n Command was: %s.\n Maybe missing package: %s"
                    )
                    % (
                        self.pkg_path,
                        varexpand(decomp_cmd, mydict=self.env),
                        missing_package,
                    ),
                    log_path=self.logfile,
                    background=self.background,
                    level=logging.ERROR,
                )
                self.returncode = 1
                self._async_wait()
                return

        pkg_xpak = portage.xpak.tbz2(self.pkg_path)
        pkg_xpak.scan()

        # SIGPIPE handling (128 + SIGPIPE) should be compatible with
        # assert_sigpipe_ok() that's used by the ebuild unpack() helper.
        self.args = [
            self._shell_binary,
            "-c",
            (
                "cmd0=(head -c %d -- %s) cmd1=(%s) cmd2=(tar -xp %s -C %s -f -); "
                + '"${cmd0[@]}" | "${cmd1[@]}" | "${cmd2[@]}"; '
                + "p=(${PIPESTATUS[@]}) ; for i in {0..2}; do "
                + "if [[ ${p[$i]} != 0 && ${p[$i]} != %d ]] ; then "
                + 'echo command $(eval "echo \\"\'\\${cmd$i[*]}\'\\"") '
                + "failed with status ${p[$i]} ; exit ${p[$i]} ; fi ; done; "
                + "if [ ${p[$i]} != 0 ] ; then "
                + 'echo command $(eval "echo \\"\'\\${cmd$i[*]}\'\\"") '
                + "failed with status ${p[$i]} ; exit ${p[$i]} ; fi ; "
                + "exit 0 ;"
            )
            % (
                pkg_xpak.filestat.st_size - pkg_xpak.xpaksize,
                portage._shell_quote(self.pkg_path),
                decomp_cmd,
                tar_options,
                portage._shell_quote(self.image_dir),
                128 + signal.SIGPIPE,
            ),
        ]

        SpawnProcess._start(self)



# Copyright: 2009-2020 Gentoo Authors
# Author(s): Petteri Räty (betelgeuse@gentoo.org)
# License: GPL2

__all__ = ["database"]

import errno

from portage.cache import fs_template
from portage.versions import catsplit
from portage import cpv_getkey
from portage import os
from portage import _encodings
from portage import _unicode_decode
from portage.util._xattr import xattr


class NoValueException(Exception):
    pass


class database(fs_template.FsBased):

    autocommits = True

    def __init__(self, *args, **config):
        super().__init__(*args, **config)
        self.portdir = self.label
        self.ns = xattr.NS_USER + ".gentoo.cache"
        self.keys = set(self._known_keys)
        self.keys.add("_mtime_")
        self.keys.add("_eclasses_")
        # xattrs have an upper length
        self.max_len = self.__get_max()

    def __get_max(self):
        path = os.path.join(self.portdir, "profiles/repo_name")
        try:
            return int(self.__get(path, "value_max_len"))
        except NoValueException as e:
            maxattrlength = self.__calc_max(path)
            self.__set(path, "value_max_len", str(maxattrlength))
            return maxattrlength

    def __calc_max(self, path):
        """Find out max attribute length supported by the file system"""

        hundred = ""
        for i in range(100):
            hundred += "a"

        s = hundred

        # Could use finally but needs python 2.5 then
        try:
            while True:
                self.__set(path, "test_max", s)
                s += hundred
        except OSError as e:
            # ext based give wrong errno
            # https://bugzilla.kernel.org/show_bug.cgi?id=12793
            if e.errno in (errno.E2BIG, errno.ENOSPC):
                result = len(s) - 100
            else:
                raise

        try:
            self.__remove(path, "test_max")
        except OSError as e:
            if e.errno != errno.ENODATA:
                raise

        return result

    def __get_path(self, cpv):
        cat, pn = catsplit(cpv_getkey(cpv))
        return os.path.join(self.portdir, cat, pn, os.path.basename(cpv) + ".ebuild")

    def __has_cache(self, path):
        try:
            self.__get(path, "_mtime_")
        except NoValueException as e:
            return False

        return True

    def __get(self, path, key, default=None):
        try:
            return xattr.get(path, key, namespace=self.ns)
        except OSError as e:
            if not default is None and errno.ENODATA == e.errno:
                return default
            raise NoValueException()

    def __remove(self, path, key):
        xattr.remove(path, key, namespace=self.ns)

    def __set(self, path, key, value):
        xattr.set(path, key, value, namespace=self.ns)

    def _getitem(self, cpv):
        values = {}
        path = self.__get_path(cpv)
        attrs = {key: value for key, value in xattr.get_all(path, namespace=self.ns)}

        if not "_mtime_" in all:
            raise KeyError(cpv)

        # We default to '' like other caches
        for key in self.keys:
            attr_value = attrs.get(key, "1:")
            parts, sep, value = attr_value.partition(":")
            parts = int(parts)
            if parts > 1:
                for i in range(1, parts):
                    value += attrs.get(key + str(i))
            values[key] = value

        return values

    def _setitem(self, cpv, values):
        path = self.__get_path(cpv)
        max_len = self.max_len
        for key, value in values.items():
            # mtime comes in as long so need to convert to strings
            s = str(value)
            # We need to split long values
            value_len = len(s)
            parts = 0
            if value_len > max_len:
                # Find out how many parts we need
                parts = value_len / max_len
                if value_len % max_len > 0:
                    parts += 1

                # Only the first entry carries the number of parts
                self.__set(path, key, "{}:{}".format(parts, s[0:max_len]))

                # Write out the rest
                for i in range(1, parts):
                    start = i * max_len
                    val = s[start : start + max_len]
                    self.__set(path, key + str(i), val)
            else:
                self.__set(path, key, "{}:{}".format(1, s))

    def _delitem(self, cpv):
        pass  # Will be gone with the ebuild

    def __contains__(self, cpv):
        return os.path.exists(self.__get_path(cpv))

    def __iter__(self):

        for root, dirs, files in os.walk(self.portdir):
            for file in files:
                try:
                    file = _unicode_decode(
                        file, encoding=_encodings["fs"], errors="strict"
                    )
                except UnicodeDecodeError:
                    continue
                if file[-7:] == ".ebuild":
                    cat = os.path.basename(os.path.dirname(root))
                    pn_pv = file[:-7]
                    path = os.path.join(root, file)
                    if self.__has_cache(path):
                        yield "{}/{}/{}".format(cat, os.path.basename(root), file[:-7])



# test_dodir.py -- Portage Unit Testing Functionality
# Copyright 2007-2010 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests.bin.setup_env import BinTestCase, dodir, exists_in_D


class DoDir(BinTestCase):
    def testDoDir(self):
        self.init()
        try:
            dodir("usr /usr")
            exists_in_D("/usr")
            dodir("boot")
            exists_in_D("/boot")
            dodir("/var/lib/moocow")
            exists_in_D("/var/lib/moocow")
        finally:
            self.cleanup()



# test_normalizePath.py -- Portage Unit Testing Functionality
# Copyright 2006-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase


class NormalizePathTestCase(TestCase):
    def testNormalizePath(self):

        from portage.util import normalize_path

        path = "///foo/bar/baz"
        good = "/foo/bar/baz"
        self.assertEqual(normalize_path(path), good)



# Copyright 2010-2014 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

__all__ = ["digraph"]

import bisect
from collections import deque

from portage.util import writemsg


class digraph:
    """
    A directed graph object.
    """

    def __init__(self):
        """Create an empty digraph"""

        # { node : ( { child : priority } , { parent : priority } ) }
        self.nodes = {}
        self.order = []

    def add(self, node, parent, priority=0):
        """Adds the specified node with the specified parent.

        If the dep is a soft-dep and the node already has a hard
        relationship to the parent, the relationship is left as hard."""

        if node not in self.nodes:
            self.nodes[node] = ({}, {}, node)
            self.order.append(node)

        if not parent:
            return

        if parent not in self.nodes:
            self.nodes[parent] = ({}, {}, parent)
            self.order.append(parent)

        priorities = self.nodes[node][1].get(parent)
        if priorities is None:
            priorities = []
            self.nodes[node][1][parent] = priorities
            self.nodes[parent][0][node] = priorities

        if not priorities or priorities[-1] is not priority:
            bisect.insort(priorities, priority)

    def discard(self, node):
        """
        Like remove(), except it doesn't raises KeyError if the
        node doesn't exist.
        """
        try:
            self.remove(node)
        except KeyError:
            pass

    def remove(self, node):
        """Removes the specified node from the digraph, also removing
        and ties to other nodes in the digraph. Raises KeyError if the
        node doesn't exist."""

        if node not in self.nodes:
            raise KeyError(node)

        for parent in self.nodes[node][1]:
            del self.nodes[parent][0][node]
        for child in self.nodes[node][0]:
            del self.nodes[child][1][node]

        del self.nodes[node]
        self.order.remove(node)

    def update(self, other):
        """
        Add all nodes and edges from another digraph instance.
        """
        for node in other.order:
            children, parents, node = other.nodes[node]
            if parents:
                for parent, priorities in parents.items():
                    for priority in priorities:
                        self.add(node, parent, priority=priority)
            else:
                self.add(node, None)

    def clear(self):
        """
        Remove all nodes and edges.
        """
        self.nodes.clear()
        del self.order[:]

    def difference_update(self, t):
        """
        Remove all given nodes from node_set. This is more efficient
        than multiple calls to the remove() method.
        """
        if isinstance(t, (list, tuple)) or not hasattr(t, "__contains__"):
            t = frozenset(t)
        order = []
        for node in self.order:
            if node not in t:
                order.append(node)
                continue
            for parent in self.nodes[node][1]:
                del self.nodes[parent][0][node]
            for child in self.nodes[node][0]:
                del self.nodes[child][1][node]
            del self.nodes[node]
        self.order = order

    def has_edge(self, child, parent):
        """
        Return True if the given edge exists.
        """
        try:
            return child in self.nodes[parent][0]
        except KeyError:
            return False

    def remove_edge(self, child, parent):
        """
        Remove edge in the direction from child to parent. Note that it is
        possible for a remaining edge to exist in the opposite direction.
        Any endpoint vertices that become isolated will remain in the graph.
        """

        # Nothing should be modified when a KeyError is raised.
        for k in parent, child:
            if k not in self.nodes:
                raise KeyError(k)

        # Make sure the edge exists.
        if child not in self.nodes[parent][0]:
            raise KeyError(child)
        if parent not in self.nodes[child][1]:
            raise KeyError(parent)

        # Remove the edge.
        del self.nodes[child][1][parent]
        del self.nodes[parent][0][child]

    def __iter__(self):
        return iter(self.order)

    def contains(self, node):
        """Checks if the digraph contains mynode"""
        return node in self.nodes

    def get(self, key, default=None):
        node_data = self.nodes.get(key, self)
        if node_data is self:
            return default
        return node_data[2]

    def all_nodes(self):
        """Return a list of all nodes in the graph"""
        return self.order[:]

    def child_nodes(self, node, ignore_priority=None):
        """Return all children of the specified node"""
        if ignore_priority is None:
            return list(self.nodes[node][0])
        children = []
        if hasattr(ignore_priority, "__call__"):
            for child, priorities in self.nodes[node][0].items():
                for priority in reversed(priorities):
                    if not ignore_priority(priority):
                        children.append(child)
                        break
        else:
            for child, priorities in self.nodes[node][0].items():
                if ignore_priority < priorities[-1]:
                    children.append(child)
        return children

    def parent_nodes(self, node, ignore_priority=None):
        """Return all parents of the specified node"""
        if ignore_priority is None:
            return list(self.nodes[node][1])
        parents = []
        if hasattr(ignore_priority, "__call__"):
            for parent, priorities in self.nodes[node][1].items():
                for priority in reversed(priorities):
                    if not ignore_priority(priority):
                        parents.append(parent)
                        break
        else:
            for parent, priorities in self.nodes[node][1].items():
                if ignore_priority < priorities[-1]:
                    parents.append(parent)
        return parents

    def leaf_nodes(self, ignore_priority=None):
        """Return all nodes that have no children

        If ignore_soft_deps is True, soft deps are not counted as
        children in calculations."""

        leaf_nodes = []
        if ignore_priority is None:
            for node in self.order:
                if not self.nodes[node][0]:
                    leaf_nodes.append(node)
        elif hasattr(ignore_priority, "__call__"):
            for node in self.order:
                is_leaf_node = True
                for child, priorities in self.nodes[node][0].items():
                    for priority in reversed(priorities):
                        if not ignore_priority(priority):
                            is_leaf_node = False
                            break
                    if not is_leaf_node:
                        break
                if is_leaf_node:
                    leaf_nodes.append(node)
        else:
            for node in self.order:
                is_leaf_node = True
                for child, priorities in self.nodes[node][0].items():
                    if ignore_priority < priorities[-1]:
                        is_leaf_node = False
                        break
                if is_leaf_node:
                    leaf_nodes.append(node)
        return leaf_nodes

    def root_nodes(self, ignore_priority=None):
        """Return all nodes that have no parents.

        If ignore_soft_deps is True, soft deps are not counted as
        parents in calculations."""

        root_nodes = []
        if ignore_priority is None:
            for node in self.order:
                if not self.nodes[node][1]:
                    root_nodes.append(node)
        elif hasattr(ignore_priority, "__call__"):
            for node in self.order:
                is_root_node = True
                for parent, priorities in self.nodes[node][1].items():
                    for priority in reversed(priorities):
                        if not ignore_priority(priority):
                            is_root_node = False
                            break
                    if not is_root_node:
                        break
                if is_root_node:
                    root_nodes.append(node)
        else:
            for node in self.order:
                is_root_node = True
                for parent, priorities in self.nodes[node][1].items():
                    if ignore_priority < priorities[-1]:
                        is_root_node = False
                        break
                if is_root_node:
                    root_nodes.append(node)
        return root_nodes

    def __bool__(self):
        return bool(self.nodes)

    def is_empty(self):
        """Checks if the digraph is empty"""
        return len(self.nodes) == 0

    def clone(self):
        clone = digraph()
        clone.nodes = {}
        memo = {}
        for children, parents, node in self.nodes.values():
            children_clone = {}
            for child, priorities in children.items():
                priorities_clone = memo.get(id(priorities))
                if priorities_clone is None:
                    priorities_clone = priorities[:]
                    memo[id(priorities)] = priorities_clone
                children_clone[child] = priorities_clone
            parents_clone = {}
            for parent, priorities in parents.items():
                priorities_clone = memo.get(id(priorities))
                if priorities_clone is None:
                    priorities_clone = priorities[:]
                    memo[id(priorities)] = priorities_clone
                parents_clone[parent] = priorities_clone
            clone.nodes[node] = (children_clone, parents_clone, node)
        clone.order = self.order[:]
        return clone

    def delnode(self, node):
        try:
            self.remove(node)
        except KeyError:
            pass

    def firstzero(self):
        leaf_nodes = self.leaf_nodes()
        if leaf_nodes:
            return leaf_nodes[0]
        return None

    def hasallzeros(self, ignore_priority=None):
        return len(self.leaf_nodes(ignore_priority=ignore_priority)) == len(self.order)

    def debug_print(self):
        def output(s):
            writemsg(s, noiselevel=-1)

        for node in self.nodes:
            output("{} ".format(node))
            if self.nodes[node][0]:
                output("depends on\n")
            else:
                output("(no children)\n")
            for child, priorities in self.nodes[node][0].items():
                output(
                    "  %s (%s)\n"
                    % (
                        child,
                        priorities[-1],
                    )
                )

    def bfs(self, start, ignore_priority=None):
        if start not in self:
            raise KeyError(start)

        queue, enqueued = deque([(None, start)]), {start}
        while queue:
            parent, n = queue.popleft()
            yield parent, n
            new = set(self.child_nodes(n, ignore_priority)) - enqueued
            enqueued |= new
            queue.extend([(n, child) for child in new])

    def shortest_path(self, start, end, ignore_priority=None):
        if start not in self:
            raise KeyError(start)
        elif end not in self:
            raise KeyError(end)

        paths = {None: []}
        for parent, child in self.bfs(start, ignore_priority):
            paths[child] = paths[parent] + [child]
            if child == end:
                return paths[child]
        return None

    def get_cycles(self, ignore_priority=None, max_length=None):
        """
        Returns all cycles that have at most length 'max_length'.
        If 'max_length' is 'None', all cycles are returned.
        """
        all_cycles = []
        for node in self.nodes:
            # If we have multiple paths of the same length, we have to
            # return them all, so that we always get the same results
            # even with PYTHONHASHSEED="random" enabled.
            shortest_path = None
            candidates = []
            for child in self.child_nodes(node, ignore_priority):
                path = self.shortest_path(child, node, ignore_priority)
                if path is None:
                    continue
                if not shortest_path or len(shortest_path) >= len(path):
                    shortest_path = path
                    candidates.append(path)
            if shortest_path and (not max_length or len(shortest_path) <= max_length):
                for path in candidates:
                    if len(path) == len(shortest_path):
                        all_cycles.append(path)
        return all_cycles

    # Backward compatibility
    addnode = add
    allnodes = all_nodes
    allzeros = leaf_nodes
    hasnode = contains
    __contains__ = contains
    empty = is_empty
    copy = clone



# Copyright 2015 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class SonameDepcleanTestCase(TestCase):
    def testSonameDepclean(self):

        installed = {
            "app-misc/A-1": {
                "RDEPEND": "dev-libs/B",
                "DEPEND": "dev-libs/B",
                "REQUIRES": "x86_32: libB.so.1 libc.so.6",
            },
            "dev-libs/B-1": {
                "PROVIDES": "x86_32: libB.so.1",
            },
            "sys-libs/glibc-2.19-r1": {"PROVIDES": "x86_32: libc.so.6"},
        }

        world = ("app-misc/A",)

        test_cases = (
            ResolverPlaygroundTestCase(
                [],
                options={
                    "--depclean": True,
                    "--ignore-soname-deps": "n",
                },
                success=True,
                cleanlist=[],
            ),
            ResolverPlaygroundTestCase(
                [],
                options={
                    "--depclean": True,
                    "--ignore-soname-deps": "y",
                },
                success=True,
                cleanlist=["sys-libs/glibc-2.19-r1"],
            ),
        )

        playground = ResolverPlayground(debug=False, installed=installed, world=world)
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            # Disable debug so that cleanup works.
            playground.debug = False
            playground.cleanup()



# Copyright 2010-2012 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

__all__ = ["digestcheck"]

import warnings

from portage import os, _encodings, _unicode_decode
from portage.checksum import _hash_filter
from portage.exception import DigestException, FileNotFound
from portage.localization import _
from portage.output import EOutput
from portage.util import writemsg


def digestcheck(myfiles, mysettings, strict=False, justmanifest=None, mf=None):
    """
    Verifies checksums. Assumes all files have been downloaded.
    @rtype: int
    @return: 1 on success and 0 on failure
    """

    if justmanifest is not None:
        warnings.warn(
            "The justmanifest parameter of the "
            + "portage.package.ebuild.digestcheck.digestcheck()"
            + " function is now unused.",
            DeprecationWarning,
            stacklevel=2,
        )
        justmanifest = None

    if mysettings.get("EBUILD_SKIP_MANIFEST") == "1":
        return 1
    pkgdir = mysettings["O"]
    hash_filter = _hash_filter(mysettings.get("PORTAGE_CHECKSUM_FILTER", ""))
    if hash_filter.transparent:
        hash_filter = None
    if mf is None:
        mf = mysettings.repositories.get_repo_for_location(
            os.path.dirname(os.path.dirname(pkgdir))
        )
        mf = mf.load_manifest(pkgdir, mysettings["DISTDIR"])
    eout = EOutput()
    eout.quiet = mysettings.get("PORTAGE_QUIET", None) == "1"
    try:
        if not mf.thin and strict and "PORTAGE_PARALLEL_FETCHONLY" not in mysettings:
            if mf.fhashdict.get("EBUILD"):
                eout.ebegin(_("checking ebuild checksums ;-)"))
                mf.checkTypeHashes("EBUILD", hash_filter=hash_filter)
                eout.eend(0)
            if mf.fhashdict.get("AUX"):
                eout.ebegin(_("checking auxfile checksums ;-)"))
                mf.checkTypeHashes("AUX", hash_filter=hash_filter)
                eout.eend(0)
            if mf.strict_misc_digests and mf.fhashdict.get("MISC"):
                eout.ebegin(_("checking miscfile checksums ;-)"))
                mf.checkTypeHashes(
                    "MISC", ignoreMissingFiles=True, hash_filter=hash_filter
                )
                eout.eend(0)
        for f in myfiles:
            eout.ebegin(_("checking %s ;-)") % f)
            ftype = mf.findFile(f)
            if ftype is None:
                if mf.allow_missing:
                    continue
                eout.eend(1)
                writemsg(_("\n!!! Missing digest for '%s'\n") % (f,), noiselevel=-1)
                return 0
            mf.checkFileHashes(ftype, f, hash_filter=hash_filter)
            eout.eend(0)
    except FileNotFound as e:
        eout.eend(1)
        writemsg(
            _("\n!!! A file listed in the Manifest could not be found: %s\n") % str(e),
            noiselevel=-1,
        )
        return 0
    except DigestException as e:
        eout.eend(1)
        writemsg(_("\n!!! Digest verification failed:\n"), noiselevel=-1)
        writemsg("!!! %s\n" % e.value[0], noiselevel=-1)
        writemsg(_("!!! Reason: %s\n") % e.value[1], noiselevel=-1)
        writemsg(_("!!! Got: %s\n") % e.value[2], noiselevel=-1)
        writemsg(_("!!! Expected: %s\n") % e.value[3], noiselevel=-1)
        return 0
    if mf.thin or mf.allow_missing:
        # In this case we ignore any missing digests that
        # would otherwise be detected below.
        return 1
    # Make sure that all of the ebuilds are actually listed in the Manifest.
    for f in os.listdir(pkgdir):
        pf = None
        if f[-7:] == ".ebuild":
            pf = f[:-7]
        if pf is not None and not mf.hasFile("EBUILD", f):
            writemsg(
                _("!!! A file is not listed in the Manifest: '%s'\n")
                % os.path.join(pkgdir, f),
                noiselevel=-1,
            )
            if strict:
                return 0
    # epatch will just grab all the patches out of a directory, so we have to
    # make sure there aren't any foreign files that it might grab.
    filesdir = os.path.join(pkgdir, "files")

    for parent, dirs, files in os.walk(filesdir):
        try:
            parent = _unicode_decode(parent, encoding=_encodings["fs"], errors="strict")
        except UnicodeDecodeError:
            parent = _unicode_decode(
                parent, encoding=_encodings["fs"], errors="replace"
            )
            writemsg(
                _("!!! Path contains invalid " "character(s) for encoding '%s': '%s'")
                % (_encodings["fs"], parent),
                noiselevel=-1,
            )
            if strict:
                return 0
            continue
        for d in dirs:
            d_bytes = d
            try:
                d = _unicode_decode(d, encoding=_encodings["fs"], errors="strict")
            except UnicodeDecodeError:
                d = _unicode_decode(d, encoding=_encodings["fs"], errors="replace")
                writemsg(
                    _(
                        "!!! Path contains invalid "
                        "character(s) for encoding '%s': '%s'"
                    )
                    % (_encodings["fs"], os.path.join(parent, d)),
                    noiselevel=-1,
                )
                if strict:
                    return 0
                dirs.remove(d_bytes)
                continue
            if d.startswith(".") or d == "CVS":
                dirs.remove(d_bytes)
        for f in files:
            try:
                f = _unicode_decode(f, encoding=_encodings["fs"], errors="strict")
            except UnicodeDecodeError:
                f = _unicode_decode(f, encoding=_encodings["fs"], errors="replace")
                if f.startswith("."):
                    continue
                f = os.path.join(parent, f)[len(filesdir) + 1 :]
                writemsg(
                    _(
                        "!!! File name contains invalid "
                        "character(s) for encoding '%s': '%s'"
                    )
                    % (_encodings["fs"], f),
                    noiselevel=-1,
                )
                if strict:
                    return 0
                continue
            if f.startswith("."):
                continue
            f = os.path.join(parent, f)[len(filesdir) + 1 :]
            file_type = mf.findFile(f)
            if file_type != "AUX" and not f.startswith("digest-"):
                writemsg(
                    _("!!! A file is not listed in the Manifest: '%s'\n")
                    % os.path.join(filesdir, f),
                    noiselevel=-1,
                )
                if strict:
                    return 0
    return 1



#!/usr/bin/env python
# Copyright 2010-2022 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2
#
# This is a helper which ebuild processes can use
# to communicate with portage's main python process.

# This block ensures that ^C interrupts are handled quietly.
try:
    import os
    import signal

    def exithandler(signum, _frame):
        signal.signal(signum, signal.SIG_DFL)
        os.kill(os.getpid(), signum)

    signal.signal(signal.SIGINT, exithandler)
    signal.signal(signal.SIGTERM, exithandler)
    signal.signal(signal.SIGPIPE, signal.SIG_DFL)

except KeyboardInterrupt:
    raise SystemExit(130)

import errno
import logging
import pickle
import sys
import time


def debug_signal(signum, frame):
    import pdb

    pdb.set_trace()


signal.signal(signal.SIGUSR1, debug_signal)

if os.path.isfile(
    os.path.join(
        os.path.dirname(os.path.dirname(os.path.realpath(__file__))),
        ".portage_not_installed",
    )
):
    pym_paths = [
        os.path.join(
            os.path.dirname(os.path.dirname(os.path.realpath(__file__))), "lib"
        )
    ]
    sys.path.insert(0, pym_paths[0])
else:
    import sysconfig

    pym_paths = [
        os.path.join(sysconfig.get_path("purelib"), x) for x in ("_emerge", "portage")
    ]
# Avoid sandbox violations after Python upgrade.
if os.environ.get("SANDBOX_ON") == "1":
    sandbox_write = os.environ.get("SANDBOX_WRITE", "").split(":")
    for pym_path in pym_paths:
        if pym_path not in sandbox_write:
            sandbox_write.append(pym_path)
            os.environ["SANDBOX_WRITE"] = ":".join(filter(None, sandbox_write))
    del pym_path, sandbox_write
del pym_paths

import portage

portage._internal_caller = True
portage._disable_legacy_globals()

from portage.util._eventloop.global_event_loop import global_event_loop
from _emerge.AbstractPollTask import AbstractPollTask
from _emerge.PipeReader import PipeReader

RETURNCODE_WRITE_FAILED = 2


class FifoWriter(AbstractPollTask):

    __slots__ = ("buf", "fifo", "_fd")

    def _start(self):
        try:
            self._fd = os.open(self.fifo, os.O_WRONLY | os.O_NONBLOCK)
        except OSError as e:
            if e.errno == errno.ENXIO:
                # This happens if the daemon has been killed.
                self.returncode = RETURNCODE_WRITE_FAILED
                self._unregister()
                self._async_wait()
                return
            else:
                raise
        self.scheduler.add_writer(self._fd, self._output_handler)
        self._registered = True

    def _output_handler(self):
        # The whole buf should be able to fit in the fifo with
        # a single write call, so there's no valid reason for
        # os.write to raise EAGAIN here.
        fd = self._fd
        buf = self.buf
        while buf:
            try:
                buf = buf[os.write(fd, buf) :]
            except OSError:
                self.returncode = RETURNCODE_WRITE_FAILED
                self._async_wait()
                return

        self.returncode = os.EX_OK
        self._async_wait()

    def _cancel(self):
        self.returncode = self._cancelled_returncode
        self._unregister()

    def _unregister(self):
        self._registered = False
        if self._fd is not None:
            self.scheduler.remove_writer(self._fd)
            os.close(self._fd)
            self._fd = None


class EbuildIpc:

    # Timeout for each individual communication attempt (we retry
    # as long as the daemon process appears to be alive).
    _COMMUNICATE_RETRY_TIMEOUT = 15  # seconds

    def __init__(self):
        self.fifo_dir = os.environ["PORTAGE_BUILDDIR"]
        self.ipc_in_fifo = os.path.join(self.fifo_dir, ".ipc", "in")
        self.ipc_out_fifo = os.path.join(self.fifo_dir, ".ipc", "out")
        self.ipc_lock_file = os.path.join(self.fifo_dir, ".ipc", "lock")

    def _daemon_is_alive(self):
        try:
            builddir_lock = portage.locks.lockfile(
                self.fifo_dir, wantnewlockfile=True, flags=os.O_NONBLOCK
            )
        except portage.exception.TryAgain:
            return True
        else:
            portage.locks.unlockfile(builddir_lock)
            return False

    def communicate(self, args):

        # Make locks quiet since unintended locking messages displayed on
        # stdout could corrupt the intended output of this program.
        portage.locks._quiet = True
        lock_obj = portage.locks.lockfile(self.ipc_lock_file, unlinkfile=True)

        try:
            return self._communicate(args)
        finally:
            portage.locks.unlockfile(lock_obj)

    def _timeout_retry_msg(self, start_time, when):
        time_elapsed = time.time() - start_time
        portage.util.writemsg_level(
            portage.localization._(
                "ebuild-ipc timed out %s after %d seconds," + " retrying...\n"
            )
            % (when, time_elapsed),
            level=logging.ERROR,
            noiselevel=-1,
        )

    def _no_daemon_msg(self):
        portage.util.writemsg_level(
            portage.localization._("ebuild-ipc: daemon process not detected\n"),
            level=logging.ERROR,
            noiselevel=-1,
        )

    def _run_writer(self, fifo_writer, msg):
        """
        Wait on pid and return an appropriate exit code. This
        may return unsuccessfully due to timeout if the daemon
        process does not appear to be alive.
        """

        start_time = time.time()

        fifo_writer.start()
        eof = fifo_writer.poll() is not None

        while not eof:
            fifo_writer._wait_loop(timeout=self._COMMUNICATE_RETRY_TIMEOUT)

            eof = fifo_writer.poll() is not None
            if eof:
                break
            elif self._daemon_is_alive():
                self._timeout_retry_msg(start_time, msg)
            else:
                fifo_writer.cancel()
                self._no_daemon_msg()
                fifo_writer.wait()
                return 2

        return fifo_writer.wait()

    def _receive_reply(self, input_fd):

        start_time = time.time()

        pipe_reader = PipeReader(
            input_files={"input_fd": input_fd}, scheduler=global_event_loop()
        )
        pipe_reader.start()

        eof = pipe_reader.poll() is not None

        while not eof:
            pipe_reader._wait_loop(timeout=self._COMMUNICATE_RETRY_TIMEOUT)
            eof = pipe_reader.poll() is not None
            if not eof:
                if self._daemon_is_alive():
                    self._timeout_retry_msg(
                        start_time, portage.localization._("during read")
                    )
                else:
                    pipe_reader.cancel()
                    self._no_daemon_msg()
                    return 2

        buf = pipe_reader.getvalue()

        retval = 2

        if not buf:

            portage.util.writemsg_level(
                "ebuild-ipc: {}\n".format(portage.localization._("read failed")),
                level=logging.ERROR,
                noiselevel=-1,
            )

        else:

            try:
                reply = pickle.loads(buf)
            except SystemExit:
                raise
            except Exception as e:
                # The pickle module can raise practically
                # any exception when given corrupt data.
                portage.util.writemsg_level(
                    "ebuild-ipc: {}\n".format(e), level=logging.ERROR, noiselevel=-1
                )

            else:

                (out, err, retval) = reply

                if out:
                    portage.util.writemsg_stdout(out, noiselevel=-1)

                if err:
                    portage.util.writemsg(err, noiselevel=-1)

        return retval

    def _communicate(self, args):

        if not self._daemon_is_alive():
            self._no_daemon_msg()
            return 2

        # Open the input fifo before the output fifo, in order to make it
        # possible for the daemon to send a reply without blocking. This
        # improves performance, and also makes it possible for the daemon
        # to do a non-blocking write without a race condition.
        input_fd = os.open(self.ipc_out_fifo, os.O_RDONLY | os.O_NONBLOCK)

        # Use forks so that the child process can handle blocking IO
        # un-interrupted, while the parent handles all timeout
        # considerations. This helps to avoid possible race conditions
        # from interference between timeouts and blocking IO operations.
        msg = portage.localization._("during write")
        retval = self._run_writer(
            FifoWriter(
                buf=pickle.dumps(args),
                fifo=self.ipc_in_fifo,
                scheduler=global_event_loop(),
            ),
            msg,
        )

        if retval != os.EX_OK:
            portage.util.writemsg_level(
                "ebuild-ipc: %s: %s\n"
                % (msg, portage.localization._("subprocess failure: %s") % retval),
                level=logging.ERROR,
                noiselevel=-1,
            )
            return retval

        if not self._daemon_is_alive():
            self._no_daemon_msg()
            return 2

        return self._receive_reply(input_fd)


def ebuild_ipc_main(args):
    ebuild_ipc = EbuildIpc()
    return ebuild_ipc.communicate(args)


if __name__ == "__main__":
    try:
        sys.exit(ebuild_ipc_main(sys.argv[1:]))
    finally:
        global_event_loop().close()



# Copyright 2015-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2


class SonameAtom:

    __slots__ = ("multilib_category", "soname", "_hash_key", "_hash_value")

    # Distiguishes package atoms from other atom types
    package = False

    def __init__(self, multilib_category, soname):
        object.__setattr__(self, "multilib_category", multilib_category)
        object.__setattr__(self, "soname", soname)
        object.__setattr__(self, "_hash_key", (multilib_category, soname))
        object.__setattr__(self, "_hash_value", hash(self._hash_key))

    def __setattr__(self, name, value):
        raise AttributeError(
            "SonameAtom instances are immutable", self.__class__, name, value
        )

    def __getstate__(self):
        return {k: getattr(self, k) for k in self.__slots__}

    def __setstate__(self, state):
        for k, v in state.items():
            object.__setattr__(self, k, v)

    def __hash__(self):
        return self._hash_value

    def __eq__(self, other):
        try:
            return self._hash_key == other._hash_key
        except AttributeError:
            return False

    def __ne__(self, other):
        try:
            return self._hash_key != other._hash_key
        except AttributeError:
            return True

    def __repr__(self):
        return "{}('{}', '{}')".format(
            self.__class__.__name__,
            self.multilib_category,
            self.soname,
        )

    def __str__(self):
        return "{}: {}".format(self.multilib_category, self.soname)

    def match(self, pkg):
        """
        Check if the given package instance matches this atom. Unbuilt
        ebuilds, which do not have soname metadata, will never match.

        @param pkg: a Package instance
        @type pkg: Package
        @return: True if this atom matches pkg, otherwise False
        @rtype: bool
        """
        return pkg.provides is not None and self in pkg.provides



# Copyright 2015 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import sys

from portage.const import SUPPORTED_GENTOO_BINPKG_FORMATS
from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)
from portage.output import colorize


class SonameSlotConflictReinstallTestCase(TestCase):
    def testSonameSlotConflictReinstall(self):

        binpkgs = {
            "app-misc/A-1": {
                "PROVIDES": "x86_32: libA-1.so",
            },
            "app-misc/A-2": {
                "PROVIDES": "x86_32: libA-2.so",
            },
            "app-misc/B-0": {
                "DEPEND": "app-misc/A",
                "RDEPEND": "app-misc/A",
                "REQUIRES": "x86_32: libA-2.so",
            },
            "app-misc/C-0": {
                "EAPI": "5",
                "DEPEND": "<app-misc/A-2",
                "RDEPEND": "<app-misc/A-2",
            },
            "app-misc/D-1": {
                "PROVIDES": "x86_32: libD-1.so",
            },
            "app-misc/D-2": {
                "PROVIDES": "x86_32: libD-2.so",
            },
            "app-misc/E-0": {
                "DEPEND": "app-misc/D",
                "RDEPEND": "app-misc/D",
                "REQUIRES": "x86_32: libD-2.so",
            },
        }

        installed = {
            "app-misc/A-1": {
                "PROVIDES": "x86_32: libA-1.so",
            },
            "app-misc/B-0": {
                "DEPEND": "app-misc/A",
                "RDEPEND": "app-misc/A",
                "REQUIRES": "x86_32: libA-1.so",
            },
            "app-misc/C-0": {"DEPEND": "<app-misc/A-2", "RDEPEND": "<app-misc/A-2"},
            "app-misc/D-1": {
                "PROVIDES": "x86_32: libD-1.so",
            },
            "app-misc/E-0": {
                "DEPEND": "app-misc/D",
                "RDEPEND": "app-misc/D",
                "REQUIRES": "x86_32: libD-1.so",
            },
        }

        world = ["app-misc/B", "app-misc/C", "app-misc/E"]

        test_cases = (
            # Test bug #439688, where a slot conflict prevents an
            # upgrade and we don't want to trigger unnecessary rebuilds.
            ResolverPlaygroundTestCase(
                ["@world"],
                options={
                    "--deep": True,
                    "--ignore-soname-deps": "n",
                    "--update": True,
                    "--usepkgonly": True,
                    "--backtrack": 10,
                },
                success=True,
                mergelist=["[binary]app-misc/D-2", "[binary]app-misc/E-0"],
            ),
        )

        for binpkg_format in SUPPORTED_GENTOO_BINPKG_FORMATS:
            with self.subTest(binpkg_format=binpkg_format):
                print(colorize("HILITE", binpkg_format), end=" ... ")
                sys.stdout.flush()
                playground = ResolverPlayground(
                    binpkgs=binpkgs,
                    installed=installed,
                    world=world,
                    debug=False,
                    user_config={
                        "make.conf": ('BINPKG_FORMAT="%s"' % binpkg_format,),
                    },
                )
                try:
                    for test_case in test_cases:
                        playground.run_TestCase(test_case)
                        self.assertEqual(
                            test_case.test_success, True, test_case.fail_msg
                        )
                finally:
                    playground.debug = False
                    playground.cleanup()

    def testSonameSlotConflictMassRebuild(self):
        """
        Bug 486580
        Before this bug was fixed, emerge would backtrack for each
        package that needs a rebuild. This could cause it to hit the
        backtrack limit and not rebuild all needed packages.
        """
        binpkgs = {
            "app-misc/A-1": {
                "DEPEND": "app-misc/B",
                "RDEPEND": "app-misc/B",
                "REQUIRES": "x86_32: libB-2.so",
            },
            "app-misc/B-1": {
                "SLOT": "1",
                "PROVIDES": "x86_32: libB-1.so",
            },
            "app-misc/B-2": {
                "SLOT": "2",
                "PROVIDES": "x86_32: libB-2.so",
            },
        }

        installed = {
            "app-misc/B-1": {
                "SLOT": "1",
                "PROVIDES": "x86_32: libB-1.so",
            },
        }

        expected_mergelist = ["[binary]app-misc/A-1", "[binary]app-misc/B-2"]

        for i in range(5):
            binpkgs["app-misc/C%sC-1" % i] = {
                "DEPEND": "app-misc/B",
                "RDEPEND": "app-misc/B",
                "REQUIRES": "x86_32: libB-2.so",
            }

            installed["app-misc/C%sC-1" % i] = {
                "DEPEND": "app-misc/B",
                "RDEPEND": "app-misc/B",
                "REQUIRES": "x86_32: libB-1.so",
            }
            for x in ("DEPEND", "RDEPEND"):
                binpkgs["app-misc/A-1"][x] += " app-misc/C%sC" % i

            expected_mergelist.append("[binary]app-misc/C%sC-1" % i)

        test_cases = (
            ResolverPlaygroundTestCase(
                ["app-misc/A"],
                ignore_mergelist_order=True,
                all_permutations=True,
                options={
                    "--backtrack": 3,
                    "--deep": True,
                    "--ignore-soname-deps": "n",
                    "--update": True,
                    "--usepkgonly": True,
                },
                success=True,
                mergelist=expected_mergelist,
            ),
        )

        world = []

        for binpkg_format in SUPPORTED_GENTOO_BINPKG_FORMATS:
            with self.subTest(binpkg_format=binpkg_format):
                print(colorize("HILITE", binpkg_format), end=" ... ")
                sys.stdout.flush()
                playground = ResolverPlayground(
                    binpkgs=binpkgs,
                    installed=installed,
                    world=world,
                    debug=False,
                    user_config={
                        "make.conf": ('BINPKG_FORMAT="%s"' % binpkg_format,),
                    },
                )

                try:
                    for test_case in test_cases:
                        playground.run_TestCase(test_case)
                        self.assertEqual(
                            test_case.test_success, True, test_case.fail_msg
                        )
                finally:
                    playground.debug = False
                    playground.cleanup()

    def testSonameSlotConflictForgottenChild(self):
        """
        Similar to testSonameSlotConflictMassRebuild above, but this
        time the rebuilds are scheduled, but the package causing the
        rebuild (the child) is not installed.
        """
        binpkgs = {
            "app-misc/A-2": {
                "DEPEND": "app-misc/B app-misc/C",
                "RDEPEND": "app-misc/B app-misc/C",
                "REQUIRES": "x86_32: libB-2.so",
            },
            "app-misc/B-2": {
                "PROVIDES": "x86_32: libB-2.so",
                "SLOT": "2",
            },
            "app-misc/C-1": {
                "DEPEND": "app-misc/B",
                "RDEPEND": "app-misc/B",
                "REQUIRES": "x86_32: libB-2.so",
            },
        }

        installed = {
            "app-misc/A-1": {
                "DEPEND": "app-misc/B app-misc/C",
                "RDEPEND": "app-misc/B app-misc/C",
                "REQUIRES": "x86_32: libB-1.so",
            },
            "app-misc/B-1": {
                "PROVIDES": "x86_32: libB-1.so",
                "SLOT": "1",
            },
            "app-misc/C-1": {
                "DEPEND": "app-misc/B",
                "RDEPEND": "app-misc/B",
                "REQUIRES": "x86_32: libB-1.so",
            },
        }

        test_cases = (
            ResolverPlaygroundTestCase(
                ["app-misc/A"],
                options={
                    "--ignore-soname-deps": "n",
                    "--usepkgonly": True,
                },
                success=True,
                mergelist=[
                    "[binary]app-misc/B-2",
                    "[binary]app-misc/A-2",
                ],
            ),
            ResolverPlaygroundTestCase(
                ["@world"],
                options={
                    "--ignore-soname-deps": "n",
                    "--usepkgonly": True,
                    "--update": True,
                    "--deep": True,
                },
                success=True,
                mergelist=[
                    "[binary]app-misc/B-2",
                    "[binary]app-misc/C-1",
                    "[binary]app-misc/A-2",
                ],
            ),
        )

        world = ["app-misc/A"]

        for binpkg_format in SUPPORTED_GENTOO_BINPKG_FORMATS:
            with self.subTest(binpkg_format=binpkg_format):
                print(colorize("HILITE", binpkg_format), end=" ... ")
                sys.stdout.flush()
                playground = ResolverPlayground(
                    binpkgs=binpkgs,
                    installed=installed,
                    world=world,
                    debug=False,
                    user_config={
                        "make.conf": ('BINPKG_FORMAT="%s"' % binpkg_format,),
                    },
                )

                try:
                    for test_case in test_cases:
                        playground.run_TestCase(test_case)
                        self.assertEqual(
                            test_case.test_success, True, test_case.fail_msg
                        )
                finally:
                    playground.debug = False
                    playground.cleanup()

    def testSonameSlotConflictMixedDependencies(self):
        """
        Bug 487198
        For parents with mixed >= and < dependencies, we scheduled
        reinstalls for the >= atom, but in the end didn't install the
        child update because of the < atom.
        """
        binpkgs = {
            "cat/slotted-lib-1": {
                "PROVIDES": "x86_32: lib1.so",
                "SLOT": "1",
            },
            "cat/slotted-lib-2": {
                "PROVIDES": "x86_32: lib2.so",
                "SLOT": "2",
            },
            "cat/slotted-lib-3": {
                "PROVIDES": "x86_32: lib3.so",
                "SLOT": "3",
            },
            "cat/slotted-lib-4": {
                "PROVIDES": "x86_32: lib4.so",
                "SLOT": "4",
            },
            "cat/slotted-lib-5": {
                "PROVIDES": "x86_32: lib5.so",
                "SLOT": "5",
            },
            "cat/user-1": {
                "DEPEND": ">=cat/slotted-lib-2 <cat/slotted-lib-4",
                "RDEPEND": ">=cat/slotted-lib-2 <cat/slotted-lib-4",
                "REQUIRES": "x86_32: lib3.so",
            },
        }

        installed = {
            "cat/slotted-lib-3": {
                "PROVIDES": "x86_32: lib3.so",
                "SLOT": "3",
            },
            "cat/user-1": {
                "DEPEND": ">=cat/slotted-lib-2 <cat/slotted-lib-4",
                "RDEPEND": ">=cat/slotted-lib-2 <cat/slotted-lib-4",
                "REQUIRES": "x86_32: lib3.so",
            },
        }

        test_cases = (
            ResolverPlaygroundTestCase(
                ["cat/user"],
                options={
                    "--deep": True,
                    "--ignore-soname-deps": "n",
                    "--update": True,
                    "--usepkgonly": True,
                },
                success=True,
                mergelist=[],
            ),
        )

        world = []

        for binpkg_format in SUPPORTED_GENTOO_BINPKG_FORMATS:
            with self.subTest(binpkg_format=binpkg_format):
                print(colorize("HILITE", binpkg_format), end=" ... ")
                sys.stdout.flush()
                playground = ResolverPlayground(
                    binpkgs=binpkgs,
                    installed=installed,
                    world=world,
                    debug=False,
                    user_config={
                        "make.conf": ('BINPKG_FORMAT="%s"' % binpkg_format,),
                    },
                )
                try:
                    for test_case in test_cases:
                        playground.run_TestCase(test_case)
                        self.assertEqual(
                            test_case.test_success, True, test_case.fail_msg
                        )
                finally:
                    playground.debug = False
                    playground.cleanup()



# Copyright 2007-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

__all__ = [
    "SETPREFIX",
    "get_boolean",
    "SetConfigError",
    "SetConfig",
    "load_default_config",
]

import portage
from portage import os
from portage import load_mod
from portage import _unicode_decode
from portage import _unicode_encode
from portage import _encodings
from portage.const import USER_CONFIG_PATH, GLOBAL_CONFIG_PATH
from portage.const import VCS_DIRS
from portage.const import _ENABLE_SET_CONFIG
from portage.exception import PackageSetNotFound
from portage.localization import _
from portage.util import writemsg_level
from portage.util.configparser import (
    SafeConfigParser,
    NoOptionError,
    ParsingError,
    read_configs,
)

SETPREFIX = "@"


def get_boolean(options, name, default):
    if not name in options:
        return default
    if options[name].lower() in ("1", "yes", "on", "true"):
        return True
    if options[name].lower() in ("0", "no", "off", "false"):
        return False
    raise SetConfigError(
        _("invalid value '%(value)s' for option '%(option)s'")
        % {"value": options[name], "option": name}
    )


class SetConfigError(Exception):
    pass


class SetConfig:
    def __init__(self, paths, settings, trees):
        self._parser = SafeConfigParser(
            defaults={
                "EPREFIX": settings["EPREFIX"],
                "EROOT": settings["EROOT"],
                "PORTAGE_CONFIGROOT": settings["PORTAGE_CONFIGROOT"],
                "ROOT": settings["ROOT"],
            }
        )

        if _ENABLE_SET_CONFIG:
            read_configs(self._parser, paths)
        else:
            self._create_default_config()

        self.errors = []
        self.psets = {}
        self.trees = trees
        self.settings = settings
        self._parsed = False
        self.active = []

    def _create_default_config(self):
        """
        Create a default hardcoded set configuration for a portage version
        that does not support set configuration files. This is only used
        in the current branch of portage if _ENABLE_SET_CONFIG is False.
        Even if it's not used in this branch, keep it here in order to
        minimize the diff between branches.

                [world]
                class = portage.sets.base.DummyPackageSet
                packages = @selected @system

                [selected]
                class = portage.sets.files.WorldSelectedSet

                [system]
                class = portage.sets.profiles.PackagesSystemSet

        """
        parser = self._parser

        parser.remove_section("world")
        parser.add_section("world")
        parser.set("world", "class", "portage.sets.base.DummyPackageSet")
        parser.set("world", "packages", "@profile @selected @system")

        parser.remove_section("profile")
        parser.add_section("profile")
        parser.set(
            "profile", "class", "portage.sets.ProfilePackageSet.ProfilePackageSet"
        )

        parser.remove_section("selected")
        parser.add_section("selected")
        parser.set("selected", "class", "portage.sets.files.WorldSelectedSet")

        parser.remove_section("selected-packages")
        parser.add_section("selected-packages")
        parser.set(
            "selected-packages", "class", "portage.sets.files.WorldSelectedPackagesSet"
        )

        parser.remove_section("selected-sets")
        parser.add_section("selected-sets")
        parser.set("selected-sets", "class", "portage.sets.files.WorldSelectedSetsSet")

        parser.remove_section("system")
        parser.add_section("system")
        parser.set("system", "class", "portage.sets.profiles.PackagesSystemSet")

        parser.remove_section("security")
        parser.add_section("security")
        parser.set("security", "class", "portage.sets.security.NewAffectedSet")

        parser.remove_section("usersets")
        parser.add_section("usersets")
        parser.set("usersets", "class", "portage.sets.files.StaticFileSet")
        parser.set("usersets", "multiset", "true")
        parser.set("usersets", "directory", "%(PORTAGE_CONFIGROOT)setc/portage/sets")
        parser.set("usersets", "world-candidate", "true")

        parser.remove_section("live-rebuild")
        parser.add_section("live-rebuild")
        parser.set("live-rebuild", "class", "portage.sets.dbapi.VariableSet")
        parser.set("live-rebuild", "variable", "PROPERTIES")
        parser.set("live-rebuild", "includes", "live")

        parser.remove_section("deprecated-live-rebuild")
        parser.add_section("deprecated-live-rebuild")
        parser.set("deprecated-live-rebuild", "class", "portage.sets.dbapi.VariableSet")
        parser.set("deprecated-live-rebuild", "variable", "INHERITED")
        parser.set(
            "deprecated-live-rebuild",
            "includes",
            " ".join(sorted(portage.const.LIVE_ECLASSES)),
        )

        parser.remove_section("module-rebuild")
        parser.add_section("module-rebuild")
        parser.set("module-rebuild", "class", "portage.sets.dbapi.OwnerSet")
        parser.set("module-rebuild", "files", "/lib/modules")

        parser.remove_section("preserved-rebuild")
        parser.add_section("preserved-rebuild")
        parser.set(
            "preserved-rebuild",
            "class",
            "portage.sets.libs.PreservedLibraryConsumerSet",
        )

        parser.remove_section("x11-module-rebuild")
        parser.add_section("x11-module-rebuild")
        parser.set("x11-module-rebuild", "class", "portage.sets.dbapi.OwnerSet")
        parser.set("x11-module-rebuild", "files", "/usr/lib*/xorg/modules")
        parser.set("x11-module-rebuild", "exclude-files", "/usr/bin/Xorg")

    def update(self, setname, options):
        parser = self._parser
        self.errors = []
        if not setname in self.psets:
            options["name"] = setname
            options["world-candidate"] = "False"

            # for the unlikely case that there is already a section with the requested setname
            import random

            while setname in parser.sections():
                setname = "%08d" % random.randint(0, 10**10)

            parser.add_section(setname)
            for k, v in options.items():
                parser.set(setname, k, v)
        else:
            section = self.psets[setname].creator
            if parser.has_option(section, "multiset") and parser.getboolean(
                section, "multiset"
            ):
                self.errors.append(
                    _(
                        "Invalid request to reconfigure set '%(set)s' generated "
                        "by multiset section '%(section)s'"
                    )
                    % {"set": setname, "section": section}
                )
                return
            for k, v in options.items():
                parser.set(section, k, v)
        self._parse(update=True)

    def _parse(self, update=False):
        if self._parsed and not update:
            return
        parser = self._parser
        for sname in parser.sections():
            # find classname for current section, default to file based sets
            if not parser.has_option(sname, "class"):
                classname = "portage._sets.files.StaticFileSet"
            else:
                classname = parser.get(sname, "class")

            if classname.startswith("portage.sets."):
                # The module has been made private, but we still support
                # the previous namespace for sets.conf entries.
                classname = classname.replace("sets", "_sets", 1)

            # try to import the specified class
            try:
                setclass = load_mod(classname)
            except (ImportError, AttributeError):
                try:
                    setclass = load_mod("portage._sets." + classname)
                except (ImportError, AttributeError):
                    self.errors.append(
                        _("Could not import '%(class)s' for section " "'%(section)s'")
                        % {"class": classname, "section": sname}
                    )
                    continue
            # prepare option dict for the current section
            optdict = {}
            for oname in parser.options(sname):
                optdict[oname] = parser.get(sname, oname)

            # create single or multiple instances of the given class depending on configuration
            if parser.has_option(sname, "multiset") and parser.getboolean(
                sname, "multiset"
            ):
                if hasattr(setclass, "multiBuilder"):
                    newsets = {}
                    try:
                        newsets = setclass.multiBuilder(
                            optdict, self.settings, self.trees
                        )
                    except SetConfigError as e:
                        self.errors.append(
                            _("Configuration error in section '%s': %s")
                            % (sname, str(e))
                        )
                        continue
                    for x in newsets:
                        if x in self.psets and not update:
                            self.errors.append(
                                _("Redefinition of set '%s' (sections: '%s', '%s')")
                                % (x, self.psets[x].creator, sname)
                            )
                        newsets[x].creator = sname
                        if parser.has_option(
                            sname, "world-candidate"
                        ) and parser.getboolean(sname, "world-candidate"):
                            newsets[x].world_candidate = True
                    self.psets.update(newsets)
                else:
                    self.errors.append(
                        _(
                            "Section '%(section)s' is configured as multiset, but '%(class)s' "
                            "doesn't support that configuration"
                        )
                        % {"section": sname, "class": classname}
                    )
                    continue
            else:
                try:
                    setname = parser.get(sname, "name")
                except NoOptionError:
                    setname = sname
                if setname in self.psets and not update:
                    self.errors.append(
                        _("Redefinition of set '%s' (sections: '%s', '%s')")
                        % (setname, self.psets[setname].creator, sname)
                    )
                if hasattr(setclass, "singleBuilder"):
                    try:
                        self.psets[setname] = setclass.singleBuilder(
                            optdict, self.settings, self.trees
                        )
                        self.psets[setname].creator = sname
                        if parser.has_option(
                            sname, "world-candidate"
                        ) and parser.getboolean(sname, "world-candidate"):
                            self.psets[setname].world_candidate = True
                    except SetConfigError as e:
                        self.errors.append(
                            _("Configuration error in section '%s': %s")
                            % (sname, str(e))
                        )
                        continue
                else:
                    self.errors.append(
                        _(
                            "'%(class)s' does not support individual set creation, section '%(section)s' "
                            "must be configured as multiset"
                        )
                        % {"class": classname, "section": sname}
                    )
                    continue
        self._parsed = True

    def getSets(self):
        self._parse()
        return self.psets.copy()

    def getSetAtoms(self, setname, ignorelist=None):
        """
        This raises PackageSetNotFound if the give setname does not exist.
        """
        self._parse()
        try:
            myset = self.psets[setname]
        except KeyError:
            raise PackageSetNotFound(setname)
        myatoms = myset.getAtoms()

        if ignorelist is None:
            ignorelist = set()

        ignorelist.add(setname)
        for n in myset.getNonAtoms():
            if n.startswith(SETPREFIX):
                s = n[len(SETPREFIX) :]
                if s in self.psets:
                    if s not in ignorelist:
                        myatoms.update(self.getSetAtoms(s, ignorelist=ignorelist))
                else:
                    raise PackageSetNotFound(s)

        return myatoms


def load_default_config(settings, trees):

    if not _ENABLE_SET_CONFIG:
        return SetConfig(None, settings, trees)

    global_config_path = GLOBAL_CONFIG_PATH
    if portage.const.EPREFIX:
        global_config_path = os.path.join(
            portage.const.EPREFIX, GLOBAL_CONFIG_PATH.lstrip(os.sep)
        )
    vcs_dirs = [_unicode_encode(x, encoding=_encodings["fs"]) for x in VCS_DIRS]

    def _getfiles():
        sets_config_paths = [
            os.path.join(global_config_path, "sets"),
            *(
                os.path.join(repo.location, "sets.conf")
                for repo in trees["porttree"].dbapi.repositories
            ),
            os.path.join(settings["PORTAGE_CONFIGROOT"], USER_CONFIG_PATH, "sets.conf"),
        ]

        for sets_config_path in sets_config_paths:
            if os.path.isdir(sets_config_path):
                for path, dirs, files in os.walk(sets_config_path):
                    dirs.sort()
                    files.sort()
                    for d in dirs:
                        if d in vcs_dirs or d.startswith(b".") or d.endswith(b"~"):
                            dirs.remove(d)
                    for f in files:
                        if not f.startswith(b".") and not f.endswith(b"~"):
                            yield os.path.join(path, f)
            elif os.path.isfile(sets_config_path):
                yield sets_config_path

    return SetConfig(_getfiles(), settings, trees)



# Copyright 2014 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class AutounmaskUseBreakageTestCase(TestCase):
    def testAutounmaskUseBreakage(self):

        ebuilds = {
            "app-misc/A-0": {
                "EAPI": "5",
                "RDEPEND": "app-misc/D[-foo]",
            },
            "app-misc/B-0": {"EAPI": "5", "RDEPEND": "app-misc/D[foo]"},
            "app-misc/C-0": {"EAPI": "5", "RDEPEND": ">=app-misc/D-1"},
            "app-misc/D-0": {"EAPI": "5", "IUSE": "foo"},
            "app-misc/D-1": {"EAPI": "5", "IUSE": "bar"},
        }

        test_cases = (
            # Bug 510270
            # _solve_non_slot_operator_slot_conflicts throws
            # IndexError: tuple index out of range
            # due to autounmask USE breakage.
            ResolverPlaygroundTestCase(
                ["app-misc/C", "app-misc/B", "app-misc/A"],
                options={"--autounmask-backtrack": "y"},
                all_permutations=True,
                success=False,
                ambiguous_slot_collision_solutions=True,
                slot_collision_solutions=[None, []],
            ),
            # With --autounmask-backtrack=y:
            # emerge: there are no ebuilds built with USE flags to satisfy "app-misc/D[foo]".
            #!!! One of the following packages is required to complete your request:
            # - app-misc/D-0::test_repo (Change USE: +foo)
            # (dependency required by "app-misc/B-0::test_repo" [ebuild])
            # (dependency required by "app-misc/B" [argument])
            # Without --autounmask-backtrack=y:
            # [ebuild  N     ] app-misc/D-0  USE="foo"
            # [ebuild  N     ] app-misc/D-1  USE="-bar"
            # [ebuild  N     ] app-misc/C-0
            # [ebuild  N     ] app-misc/B-0
            # [ebuild  N     ] app-misc/A-0
            #
            #!!! Multiple package instances within a single package slot have been pulled
            #!!! into the dependency graph, resulting in a slot conflict:
            #
            # app-misc/D:0
            #
            #  (app-misc/D-0:0/0::test_repo, ebuild scheduled for merge) pulled in by
            #    app-misc/D[-foo] required by (app-misc/A-0:0/0::test_repo, ebuild scheduled for merge)
            #               ^^^^
            #    app-misc/D[foo] required by (app-misc/B-0:0/0::test_repo, ebuild scheduled for merge)
            #               ^^^
            #
            #  (app-misc/D-1:0/0::test_repo, ebuild scheduled for merge) pulled in by
            #    >=app-misc/D-1 required by (app-misc/C-0:0/0::test_repo, ebuild scheduled for merge)
            #    ^^           ^
            #
            # The following USE changes are necessary to proceed:
            # (see "package.use" in the portage(5) man page for more details)
            # required by app-misc/B-0::test_repo
            # required by app-misc/B (argument)
            # =app-misc/D-0 foo
            # NOTE: The --autounmask-backtrack=n output is preferable here,
            # because it highlights the unsolvable dependency conflict.
            # It would be better if it eliminated the autounmask suggestion,
            # since that suggestion won't solve the conflict.
        )

        playground = ResolverPlayground(ebuilds=ebuilds, debug=False)
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()



"""WebRsync module for portage"""

import logging

import portage
from portage import os
from portage.util import writemsg_level
from portage.util.futures import asyncio
from portage.output import create_color_func

good = create_color_func("GOOD")
bad = create_color_func("BAD")
warn = create_color_func("WARN")
from portage.sync.syncbase import SyncBase

try:
    from gemato.exceptions import GematoException
    import gemato.openpgp
except ImportError:
    gemato = None


class WebRsync(SyncBase):
    """WebRSync sync class"""

    short_desc = "Perform sync operations on webrsync based repositories"

    @staticmethod
    def name():
        return "WebRSync"

    def __init__(self):
        SyncBase.__init__(self, "emerge-webrsync", ">=sys-apps/portage-2.3")

    @property
    def has_bin(self):
        if (
            self._bin_command != "emerge-delta-webrsync"
            and self.repo.module_specific_options.get(
                "sync-webrsync-delta", "false"
            ).lower()
            in ("true", "yes")
        ):
            self._bin_command = "emerge-delta-webrsync"
            self.bin_command = portage.process.find_binary(self._bin_command)
            self.bin_pkg = ">=app-portage/emerge-delta-webrsync-3.7.5"

        return super().has_bin

    def sync(self, **kwargs):
        """Sync the repository"""
        if kwargs:
            self._kwargs(kwargs)

        if not self.has_bin:
            return (1, False)

        # filter these out to prevent gpg errors
        for var in ["uid", "gid", "groups"]:
            self.spawn_kwargs.pop(var, None)

        verbose = "--verbose" in self.options["emerge_config"].opts
        quiet = "--quiet" in self.options["emerge_config"].opts
        openpgp_env = None
        try:
            if self.repo.module_specific_options.get(
                "sync-webrsync-verify-signature", "false"
            ).lower() in ("true", "yes"):

                if not self.repo.sync_openpgp_key_path:
                    writemsg_level(
                        "!!! sync-openpgp-key-path is not set\n",
                        level=logging.ERROR,
                        noiselevel=-1,
                    )
                    return (1, False)

                if not os.path.isfile(self.repo.sync_openpgp_key_path):
                    writemsg_level(
                        "!!! sync-openpgp-key-path file not found: %s\n"
                        % self.repo.sync_openpgp_key_path,
                        level=logging.ERROR,
                        noiselevel=-1,
                    )
                    return (1, False)

                if gemato is None:
                    writemsg_level(
                        "!!! Verifying against specified key requires gemato-14.5+ installed\n",
                        level=logging.ERROR,
                        noiselevel=-1,
                    )
                    return (1, False)

                openpgp_env = self._get_openpgp_env(self.repo.sync_openpgp_key_path)

                out = portage.output.EOutput(quiet=quiet)
                try:
                    out.einfo(
                        "Using keys from {}".format(self.repo.sync_openpgp_key_path)
                    )
                    with open(self.repo.sync_openpgp_key_path, "rb") as f:
                        openpgp_env.import_key(f)
                    self._refresh_keys(openpgp_env)
                    self.spawn_kwargs["env"]["PORTAGE_GPG_DIR"] = openpgp_env.home
                    self.spawn_kwargs["env"]["PORTAGE_TEMP_GPG_DIR"] = openpgp_env.home
                except (GematoException, asyncio.TimeoutError) as e:
                    writemsg_level(
                        "!!! Verification impossible due to keyring problem:\n%s\n"
                        % (e,),
                        level=logging.ERROR,
                        noiselevel=-1,
                    )
                    return (1, False)

            webrsync_cmd = [self.bin_command]
            if verbose:
                webrsync_cmd.append("-v")
            elif quiet:
                webrsync_cmd.append("-q")

            if self.repo.module_specific_options.get(
                "sync-webrsync-keep-snapshots", "false"
            ).lower() in ("true", "yes"):
                webrsync_cmd.append("-k")

            exitcode = portage.process.spawn(webrsync_cmd, **self.spawn_kwargs)
            if exitcode != os.EX_OK:
                msg = "!!! emerge-webrsync error in %s" % self.repo.location
                self.logger(self.xterm_titles, msg)
                writemsg_level(msg + "\n", level=logging.ERROR, noiselevel=-1)
                return (exitcode, False)
            return (exitcode, True)
        finally:
            if openpgp_env is not None:
                openpgp_env.close()


class PyWebRsync(SyncBase):
    """WebRSync sync class"""

    short_desc = "Perform sync operations on webrsync based repositories"

    @staticmethod
    def name():
        return "WebRSync"

    def __init__(self):
        SyncBase.__init__(self, None, ">=sys-apps/portage-2.3")

    def sync(self, **kwargs):
        """Sync the repository"""
        pass



# Copyright 1999-2011 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from _emerge.DependencyArg import DependencyArg
from _emerge.Package import Package
import portage
from portage._sets.base import InternalPackageSet
from portage.dep import _repo_separator


class PackageArg(DependencyArg):
    def __init__(self, package=None, **kwargs):
        DependencyArg.__init__(self, **kwargs)
        self.package = package
        atom = "=" + package.cpv
        if package.repo != Package.UNKNOWN_REPO:
            atom += _repo_separator + package.repo
        self.atom = portage.dep.Atom(atom, allow_repo=True)
        self.pset = InternalPackageSet(initial_atoms=(self.atom,), allow_repo=True)



# Copyright 2014-2018 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import sys

from portage.const import SUPPORTED_GENTOO_BINPKG_FORMATS
from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)
from portage.output import colorize


class SlotOperatorRebuildTestCase(TestCase):
    def testSlotOperatorRebuild(self):

        ebuilds = {
            "app-misc/A-1": {"EAPI": "5", "SLOT": "0/1"},
            "app-misc/A-2": {"EAPI": "5", "SLOT": "0/2"},
            "app-misc/B-0": {"EAPI": "5", "RDEPEND": "app-misc/A:="},
            "app-misc/C-0": {"EAPI": "5", "RDEPEND": "|| ( app-misc/X app-misc/A:= )"},
            "app-misc/D-1": {
                "EAPI": "6",
                "RDEPEND": "app-misc/E",
            },
            "app-misc/E-1": {
                "EAPI": "6",
                "RDEPEND": "app-misc/F:=",
            },
            "app-misc/F-1": {"EAPI": "6", "SLOT": "0/1"},
            "app-misc/F-2": {"EAPI": "6", "SLOT": "0/2"},
        }

        binpkgs = {
            "app-misc/E-1": {
                "EAPI": "6",
                "RDEPEND": "app-misc/F:0/1=",
            },
        }

        installed = {
            "app-misc/A-1": {"EAPI": "5", "SLOT": "0/1"},
            "app-misc/B-0": {"EAPI": "5", "RDEPEND": "app-misc/A:0/1="},
            "app-misc/C-0": {
                "EAPI": "5",
                "RDEPEND": "|| ( app-misc/X app-misc/A:0/1= )",
            },
            "app-misc/F-2": {"EAPI": "6", "SLOT": "0/2"},
        }

        world = ["app-misc/B", "app-misc/C"]

        test_cases = (
            # Test bug #522652, where the unsatisfiable app-misc/X
            # atom is selected, and the dependency is placed into
            # _initially_unsatisfied_deps where it is ignored, causing
            # the app-misc/C-0 rebuild to be missed.
            ResolverPlaygroundTestCase(
                ["app-misc/A"],
                options={"--dynamic-deps": "n"},
                success=True,
                ambiguous_merge_order=True,
                mergelist=["app-misc/A-2", ("app-misc/B-0", "app-misc/C-0")],
            ),
            # Test bug #652938, where a binary package built against an
            # older subslot triggered downgrade of an installed package.
            # In this case we want to reject the app-misc/E-1 binary
            # package, and rebuild it against the installed instance of
            # app-misc/F.
            ResolverPlaygroundTestCase(
                ["app-misc/D"],
                options={"--usepkg": True},
                success=True,
                mergelist=["app-misc/E-1", "app-misc/D-1"],
            ),
        )

        for binpkg_format in SUPPORTED_GENTOO_BINPKG_FORMATS:
            with self.subTest(binpkg_format=binpkg_format):
                print(colorize("HILITE", binpkg_format), end=" ... ")
                sys.stdout.flush()
                playground = ResolverPlayground(
                    ebuilds=ebuilds,
                    binpkgs=binpkgs,
                    installed=installed,
                    world=world,
                    debug=False,
                    user_config={
                        "make.conf": ('BINPKG_FORMAT="%s"' % binpkg_format,),
                    },
                )
                try:
                    for test_case in test_cases:
                        playground.run_TestCase(test_case)
                        self.assertEqual(
                            test_case.test_success, True, test_case.fail_msg
                        )
                finally:
                    playground.cleanup()
